"""
This test module still uses the old API. It will be removed in the future, when
lsl.GooseModel is removed.
"""

import jax
import numpy as np
import pytest
import tensorflow_probability.substrates.jax.distributions as tfd

import liesel.goose as gs
import liesel.model as lsl
from liesel.model.goose import finite_discrete_gibbs_kernel


class TestFiniteDiscreteGibbsKernel:
    def test_transition(self):
        values = [0, 1, 2]
        prior_probs = [0.1, 0.2, 0.7]
        value_grid = lsl.Var(values, name="value_grid")

        prior = lsl.Dist(tfd.FiniteDiscrete, outcomes=value_grid, probs=prior_probs)
        categorical_var = lsl.Var(
            value=values[0],
            distribution=prior,
            name="categorical_var",
        )

        model = lsl.Model([categorical_var])
        kernel = finite_discrete_gibbs_kernel("categorical_var", model)

        draw = kernel._transition_fn(jax.random.PRNGKey(0), model.state)
        assert draw["categorical_var"] == pytest.approx(1)

        draw = kernel._transition_fn(jax.random.PRNGKey(1), model.state)
        assert draw["categorical_var"] == pytest.approx(2)

        draw = kernel._transition_fn(jax.random.PRNGKey(2), model.state)
        assert draw["categorical_var"] == pytest.approx(1)

    def test_transition_jit(self):
        values = [0, 1, 2]
        prior_probs = [0.1, 0.2, 0.7]
        value_grid = lsl.Var(values, name="value_grid")

        prior = lsl.Dist(tfd.FiniteDiscrete, outcomes=value_grid, probs=prior_probs)
        categorical_var = lsl.Var(
            value=values[0],
            distribution=prior,
            name="categorical_var",
        )

        model = lsl.Model([categorical_var])
        kernel = finite_discrete_gibbs_kernel("categorical_var", model)

        draw = jax.jit(kernel._transition_fn)(jax.random.PRNGKey(1), model.state)
        assert draw["categorical_var"] == pytest.approx(2)

    @pytest.mark.mcmc
    def test_sample_categorical(self):
        values = [0.0, 1.0, 2.0]
        prior_probs = [0.1, 0.2, 0.7]
        value_grid = lsl.Var(values, name="value_grid")

        prior = lsl.Dist(tfd.FiniteDiscrete, outcomes=value_grid, probs=prior_probs)
        categorical_var = lsl.Var(
            value=values[0],
            distribution=prior,
            name="categorical_var",
        )

        model = lsl.Model([categorical_var])
        kernel = finite_discrete_gibbs_kernel("categorical_var", model)

        eb = gs.EngineBuilder(1, num_chains=1)
        eb.add_kernel(kernel)
        eb.set_model(gs.LieselInterface(model))
        eb.set_initial_values(model.state)
        eb.set_duration(warmup_duration=500, posterior_duration=2000)

        engine = eb.build()
        engine.sample_all_epochs()

        results = engine.get_results()
        samples = results.get_posterior_samples()

        _, counts = np.unique(samples["categorical_var"], return_counts=True)
        relative_freq = counts / np.sum(counts)

        assert np.allclose(relative_freq, prior_probs, atol=0.1)

    @pytest.mark.mcmc
    def test_sample_bernoulli(self):
        prior_prob = 0.7
        prior = lsl.Dist(tfd.Bernoulli, probs=lsl.Value(prior_prob))
        dummy_var = lsl.Var(
            value=1,
            distribution=prior,
            name="dummy_var",
        )

        model = lsl.Model([dummy_var])
        kernel = finite_discrete_gibbs_kernel("dummy_var", model, outcomes=[0, 1])

        eb = gs.EngineBuilder(1, num_chains=1)
        eb.add_kernel(kernel)
        eb.set_model(gs.LieselInterface(model))
        eb.set_initial_values(model.state)
        eb.set_duration(warmup_duration=500, posterior_duration=2000)

        engine = eb.build()
        engine.sample_all_epochs()

        results = engine.get_results()
        samples = results.get_posterior_samples()

        _, counts = np.unique(samples["dummy_var"], return_counts=True)
        relative_freq = counts / np.sum(counts)

        assert np.allclose(relative_freq, [1 - prior_prob, prior_prob], atol=0.1)

    @pytest.mark.mcmc
    def test_bernoulli_no_outcomes(self):
        prior_prob = 0.7
        prior = lsl.Dist(tfd.Bernoulli, probs=lsl.Value(prior_prob))
        dummy_var = lsl.Var(
            value=1,
            distribution=prior,
            name="dummy_var",
        )

        model = lsl.Model([dummy_var])
        kernel = finite_discrete_gibbs_kernel("dummy_var", model)

        eb = gs.EngineBuilder(1, num_chains=1)
        eb.add_kernel(kernel)
        eb.set_model(gs.LieselInterface(model))
        eb.set_initial_values(model.state)
        eb.set_duration(warmup_duration=500, posterior_duration=2000)

        engine = eb.build()
        engine.sample_all_epochs()

        results = engine.get_results()
        samples = results.get_posterior_samples()

        _, counts = np.unique(samples["dummy_var"], return_counts=True)
        relative_freq = counts / np.sum(counts)

        assert np.allclose(relative_freq, [1 - prior_prob, prior_prob], atol=0.1)



import jax.numpy as jnp
from pytest import approx

import liesel.distributions as lsld


def test_scalar_batch():
    distribution = lsld.GaussianCopula(0.42)

    assert distribution.event_shape.as_list() == [2]
    assert list(distribution.event_shape_tensor()) == [2]

    assert distribution.batch_shape.as_list() == []
    assert list(distribution.batch_shape_tensor()) == []

    x1 = jnp.array([0.2, 0.194])
    x2 = jnp.array([0.311, 0.756])
    x3 = jnp.array([0.76, 0.245])

    p1 = 0.3118741
    p2 = -0.1548549
    p3 = -0.2560579

    assert distribution.log_prob(x1) == approx(p1)
    assert distribution.log_prob(x2) == approx(p2)
    assert distribution.log_prob(x3) == approx(p3)

    x4 = jnp.vstack([x1, x2, x3])
    p4 = jnp.array([p1, p2, p3])

    assert distribution.log_prob(x4) == approx(p4)


def test_matrix_batch():
    distribution = lsld.GaussianCopula(jnp.broadcast_to(0.42, [3, 3]))

    assert distribution.event_shape.as_list() == [2]
    assert list(distribution.event_shape_tensor()) == [2]

    assert distribution.batch_shape.as_list() == [3, 3]
    assert list(distribution.batch_shape_tensor()) == [3, 3]

    x = jnp.array([0.2, 0.194])
    p = jnp.broadcast_to(0.3118741, [3, 3])

    assert distribution.log_prob(x) == approx(p)



from __future__ import annotations

import logging
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, ParamSpec, Protocol, assert_never

import tensorflow_probability.substrates.jax.distributions as tfd

from .builder import EngineBuilder
from .interface import LieselInterface
from .types import Array, JitterFunctions, Kernel, KeyArray

if TYPE_CHECKING:
    from liesel.model import Model, Var


logger = logging.getLogger(__name__)


@dataclass
class LieselMCMC:
    """
    Manages the setup of MCMC specifications for a Liesel model.

    Parameters
    ----------
    model
        The Liesel model object containing the variables and their inference \
        specifications.
    which
        A named inference configuration to use. If None, the default inference \
        attached to each variable is used.
    """

    model: Model
    which: str | None = None

    def get_spec(self, var: Var) -> MCMCSpec | None:
        """
        Retrieve the MCMC specification for a given variable.

        Parameters
        ----------
        var
            The model variable for which to get the MCMC specification.

        Returns
        -------
        The MCMC specification if available, otherwise None.

        Raises
        ------
        ValueError
            If the inference attached to the variable is not of type ``MCMCSpec``.
        """
        inference = var.get_inference(self.which)
        if inference is None:
            return inference

        if not isinstance(inference, MCMCSpec):
            raise ValueError(
                f"Attribute 'inference' of variable {var} is of type"
                f" {type(inference)}, but expected type '{MCMCSpec}'."
            )
        return inference

    def get_kernel_groups(self) -> dict[str, _KernelGroup]:
        """
        Collect and organize model variables into kernel groups.

        Returns
        -------
        A dictionary mapping group names or variable names to their corresponding \
        kernel group specifications.

        Raises
        ------
        ValueError
            If variables in the same kernel group have inconsistent kernels or kernel \
            arguments.
        """
        vars_ = self.model.vars
        kernel_groups: dict[str, _KernelGroup] = {}

        for name, var in vars_.items():
            inference = self.get_spec(var)

            if not inference:
                continue

            group_name = inference.kernel_group

            if group_name is None:
                kernel_groups[name] = _KernelGroup(
                    kernel=inference.kernel,
                    kwargs=inference.kernel_kwargs,
                    position_keys=[name],
                )

            elif group_name in kernel_groups:
                group = kernel_groups[group_name]
                same_kernel = group.kernel is inference.kernel
                if not same_kernel:
                    raise ValueError(
                        "Found incoherent kernel classes for kernel group"
                        f" {group_name}."
                    )

                same_kwargs = group.kwargs == inference.kernel_kwargs
                if not same_kwargs:
                    raise ValueError(
                        f"Found incoherent kwargs for kernel group {group_name}."
                    )
                group.position_keys.append(name)

            else:
                kernel_groups[group_name] = _KernelGroup(
                    kernel=inference.kernel,
                    kwargs=inference.kernel_kwargs,
                    position_keys=[name],
                )

        return kernel_groups

    def get_kernel_list(self) -> list[Kernel]:
        """
        Construct the list of MCMC kernels from kernel groups.

        Returns
        -------
        A list of initialized kernel instances ready to be added to the MCMC engine.
        """
        kernel_groups = self.get_kernel_groups()
        kernel_list = [
            g.kernel(g.position_keys, **g.kwargs)  # type: ignore
            for g in kernel_groups.values()
        ]
        return kernel_list

    def get_jitter_functions(self) -> JitterFunctions:
        """
        Collect jitter functions for model variables that define a jitter distribution.

        Returns
        -------
        A dictionary mapping variable names to their jitter application functions.
        """
        jitter_functions: JitterFunctions = {}
        for name, var in self.model.vars.items():
            inference = self.get_spec(var)
            if inference is not None and inference.jitter_dist is not None:
                jitter_functions[name] = inference.apply_jitter

        return jitter_functions

    def get_engine_builder(
        self,
        seed: int,
        num_chains: int,
        apply_jitter: bool = True,
    ) -> EngineBuilder:
        """
        Create and configure an `EngineBuilder` for MCMC sampling.

        Parameters
        ----------
        seed
            Random seed for reproducibility.
        num_chains
            Number of MCMC chains.
        apply_jitter
            Whether to apply jitter to the initial states, by default True.

        Returns
        -------
        EngineBuilder
            A configured ``EngineBuilder`` instance.
        """
        eb = EngineBuilder(seed=seed, num_chains=num_chains)
        eb.set_model(LieselInterface(self.model))
        eb.set_initial_values(self.model.state)

        for kernel in self.get_kernel_list():
            eb.add_kernel(kernel)

        if apply_jitter:
            eb.set_jitter_fns(self.get_jitter_functions())

        return eb


@dataclass
class _KernelGroup:
    kernel: Callable[..., Kernel]
    kwargs: dict[str, Any] = field(default_factory=dict)
    position_keys: list[str] = field(default_factory=list)


P = ParamSpec("P")


class KernelFactory(Protocol[P]):
    """Create a kernel instance based on the provided position keys and arguments."""

    def __call__(
        self, position_keys: list[str], *args: P.args, **kwargs: P.kwargs
    ) -> Kernel: ...


@dataclass
class MCMCSpec:
    """
    Specification for the MCMC kernel and optional jitter distribution associated with a
    model variable.

    Parameters
    ----------
    kernel
        A KernelFactory that returns a ``Kernel`` instance when provided with position
        keys and keyword arguments.
    kernel_kwargs
        Additional keyword arguments to be passed to the kernel callable.
    kernel_group
        Name of the kernel group this variable belongs to. Variables in the same group \
        must share the same kernel type and arguments.
    jitter_dist
        A TensorFlow Probability distribution used to apply random jitter to the \
        initial value of the variable.
    jitter_method
        The type of jitter to be applied. This can be one of the following:
        - `none`: No jitter is applied.
        - `additive`: Additive jitter is applied.
        - `multiplicative`: Multiplicative jitter is applied.
        - `replacement`: Value is replaced when jitter is applied.
    """

    def __post_init__(self) -> None:
        if self.jitter_method not in self._JITTER_METHODS:
            raise ValueError(
                f"Invalid jitter method: {self.jitter_method}. "
                f"Expected one of {self._JITTER_METHODS}."
            )

    _JITTER_METHODS = ["additive", "multiplicative", "replacement"]

    kernel: KernelFactory
    kernel_kwargs: dict[str, Any] = field(default_factory=dict)
    kernel_group: str | None = None
    jitter_dist: tfd.Distribution | None = None
    jitter_method: Literal["additive", "multiplicative", "replacement"] = "additive"

    def apply_jitter(self, seed: KeyArray, value: Array) -> Array:
        """
        Apply random jitter to a given value using the specified jitter distribution.

        If a jitter distribution is set, a random sample from the distribution is added
        to the original value. If no jitter distribution is set, the original value is
        returned unchanged.

        Parameters
        ----------
        seed
            A PRNG key used for random sampling.
        value
            The value to which jitter should be applied.

        Returns
        -------
        The jittered value with the same shape as the input.
        """
        if self.jitter_dist is None:
            return value

        # check compatibility of shapes
        if (
            self.jitter_dist.batch_shape + self.jitter_dist.event_shape != value.shape
        ) and (
            self.jitter_dist.batch_shape.rank + self.jitter_dist.event_shape.rank > 0
        ):
            raise ValueError(
                f"Jitter distribution shapes "
                f"(batch shape {self.jitter_dist.batch_shape} "
                f"and event shape {self.jitter_dist.event_shape}) "
                f"do not match variable shape {value.shape}."
            )
        sample_shape = (
            value.shape
            if self.jitter_dist.batch_shape + self.jitter_dist.event_shape == ()
            else ()
        )

        jitter = self.jitter_dist.sample(sample_shape=sample_shape, seed=seed)

        match self.jitter_method:
            case "additive":
                value = value + jitter
            case "multiplicative":
                value = value * jitter
            case "replacement":
                value = jitter
            case _:
                assert_never(self.jitter_method)

        return value



"""
Liesel probabilistic programming framework.
"""

from . import bijectors, distributions, goose, model
from .__version__ import __version__, __version_info__
from .logging import reset_logger, setup_logger

# because logger setup takes place after importing the submodules, it only affects
# log messages emitted at runtime
setup_logger()

__all__ = [
    "__version__",
    "__version_info__",
    "bijectors",
    "distributions",
    "goose",
    "model",
    "reset_logger",
    "setup_logger",
]



import pytest
from model_lm import run_kernel_test

import liesel.goose as gs
from liesel.goose.hmc import HMCKernelState, HMCTransitionInfo, HMCTuningInfo
from liesel.goose.types import Kernel


def type_check() -> None:
    kernel = gs.HMCKernel(["beta", "log_sigma"])
    _: Kernel[HMCKernelState, HMCTransitionInfo, HMCTuningInfo] = kernel


@pytest.mark.mcmc
def test_hmc(mcmc_seed):
    kernel = gs.HMCKernel(["beta", "log_sigma"])
    run_kernel_test(mcmc_seed, [kernel])



import jax.random as rd
import numpy as np
import pytest
import tensorflow_probability.substrates.jax.distributions as tfd

import liesel.goose as gs
from liesel.goose.types import Position
from liesel.model.model import GraphBuilder, Model
from liesel.model.nodes import Dist, Var


@pytest.fixture
def model():
    key = rd.PRNGKey(1337)
    mu = Var(0.0, name="mu")
    sigma = Var(1.0, name="sigma")

    x = Var(rd.normal(key, shape=(500,)), name="x")
    y = Var(x, Dist(tfd.Normal, loc=mu, scale=sigma), "y")
    yield Model([y])


class TestLieselInterface:
    def test_get_position(self, model) -> None:
        gm = gs.LieselInterface(model)
        pos = gm.extract_position(["mu_value"], model.state)
        assert pos["mu_value"] == 0.0

    def test_update_state(self, model) -> None:
        gm = gs.LieselInterface(model)
        pos = Position({"mu_value": 10.0})
        new_state = gm.update_state(pos, model.state)
        assert new_state["mu_value"].value == 10.0

    def test_get_log_prob(self, model) -> None:
        gm = gs.LieselInterface(model)
        lp_before = gm.log_prob(model.state)

        pos = Position({"mu_value": 10.0})
        new_state = gm.update_state(pos, model.state)
        lp_after = gm.log_prob(new_state)

        assert lp_before == pytest.approx(-719.46875)
        assert lp_after == pytest.approx(-25616.779296875)


@pytest.mark.mcmc
def test_sample_model(model) -> None:
    mcmc_seed = 1337
    builder = gs.EngineBuilder(mcmc_seed, num_chains=1)

    builder.add_kernel(gs.NUTSKernel(["mu_value"]))

    goose_model = gs.LieselInterface(model)
    builder.set_model(goose_model)

    builder.set_initial_values(model.state)

    builder.set_duration(
        warmup_duration=1000,
        posterior_duration=1000,
    )

    engine = builder.build()
    engine.sample_all_epochs()
    results = engine.get_results()
    samples = results.get_posterior_samples()
    avg_mu = np.mean(samples["mu_value"], axis=(0, 1))

    assert avg_mu == pytest.approx(0.0, abs=0.05)


@pytest.mark.mcmc
def test_sample_transformed_model(model: Model):
    _, vars = model.copy_nodes_and_vars()
    vsigma = vars["sigma"]
    vsigma.dist_node = Dist(tfd.InverseGamma, 0.1, 0.1)
    gb = GraphBuilder().add(*vars.values())
    vsigma.transform()
    model = gb.build_model()

    mcmc_seed = 1337
    builder = gs.EngineBuilder(mcmc_seed, num_chains=1)

    builder.add_kernel(gs.NUTSKernel(["mu_value", "sigma_transformed_value"]))
    builder.positions_included = ["sigma"]

    goose_model = gs.LieselInterface(model)
    builder.set_model(goose_model)

    builder.set_initial_values(model.state)

    builder.set_duration(
        warmup_duration=1000,
        posterior_duration=1000,
    )

    engine = builder.build()
    engine.sample_all_epochs()
    results = engine.get_results()
    samples = results.get_posterior_samples()
    avg_mu = np.mean(samples["mu_value"], axis=(0, 1))
    avg_sigma = np.mean(samples["sigma"], axis=(0, 1))

    assert avg_mu == pytest.approx(0.0, abs=0.05)
    assert avg_sigma == pytest.approx(1.0, abs=0.05)



import jax.numpy as jnp
from pytest import approx

import liesel.bijectors as lslb


def test_bijector():
    bijector = lslb.AlgebraicSigmoid()

    assert bijector.forward(0.0) == approx(0.0)
    assert bijector.forward(-1.0) == approx(-1.0 / jnp.sqrt(2.0))
    assert bijector.forward(1.0) == approx(1.0 / jnp.sqrt(2.0))

    assert bijector.forward(-9999.0) == approx(-1.0)
    assert bijector.forward(9999.0) == approx(1.0)

    assert bijector.inverse(bijector.forward(jnp.pi)) == approx(jnp.pi)



import jax.random
import jax.scipy.stats
import pytest
from model_lm import run_kernel_test

import liesel.goose as gs
from liesel.goose.mh_kernel import (
    MHProposal,
    MHTransitionInfo,
    MHTuningInfo,
    RWKernelState,
)
from liesel.goose.types import Kernel, KeyArray, ModelState, Position


def proposal_fn(key: KeyArray, model_state: ModelState, step_size: float) -> MHProposal:
    key0, key1 = jax.random.split(key)
    beta = model_state["beta"] + step_size * jax.random.normal(
        key0, model_state["beta"].shape
    )
    log_sigma = model_state["log_sigma"] + step_size * jax.random.normal(
        key1, model_state["log_sigma"].shape
    )
    return MHProposal(
        position=Position({"beta": beta, "log_sigma": log_sigma}), log_correction=0.0
    )


def proposal_asym_fn(
    key: KeyArray, model_state: ModelState, step_size: float
) -> MHProposal:
    key0, key1 = jax.random.split(key)

    mean_prop = model_state["beta"] + step_size
    beta = mean_prop + step_size * jax.random.normal(key0, model_state["beta"].shape)
    log_d_prop = jax.scipy.stats.norm.logpdf(beta, loc=mean_prop, scale=step_size).sum()

    mean_old = beta + step_size
    log_d_old = jax.scipy.stats.norm.logpdf(
        model_state["beta"], loc=mean_old, scale=step_size
    ).sum()

    log_sigma = model_state["log_sigma"] + step_size * jax.random.normal(
        key1, model_state["log_sigma"].shape
    )

    log_correction = log_d_old - log_d_prop

    return MHProposal(
        position=Position({"beta": beta, "log_sigma": log_sigma}),
        log_correction=log_correction,
    )


def type_check() -> None:
    kernel = gs.MHKernel(["beta", "log_sigma"], proposal_fn)
    _: Kernel[RWKernelState, MHTransitionInfo, MHTuningInfo] = kernel


@pytest.mark.mcmc
def test_mh_kernel_symmetric(mcmc_seed) -> None:
    kernel = gs.MHKernel(["beta", "log_sigma"], proposal_fn, da_tune_step_size=True)
    run_kernel_test(mcmc_seed, [kernel])


@pytest.mark.mcmc
def test_mh_kernel_asymmetric(mcmc_seed) -> None:
    kernel = gs.MHKernel(
        ["beta", "log_sigma"],
        proposal_asym_fn,
        da_tune_step_size=True,
        da_target_accept=0.1,
    )
    run_kernel_test(mcmc_seed, [kernel])



"""
some tests for the engine builder
"""

import jax
import jax.numpy as jnp
import tensorflow_probability.substrates.jax.distributions as tfd

import liesel.goose as gs
from liesel.goose.builder import EngineBuilder
from liesel.goose.interface import DictInterface


def test_seed_input():
    int_seed = 0
    key_seed = jax.random.PRNGKey(int_seed)
    builder = EngineBuilder(seed=int_seed, num_chains=2)
    builder2 = EngineBuilder(seed=key_seed, num_chains=2)

    assert jnp.all(builder._prng_key == builder2._prng_key)
    assert jnp.all(builder._engine_key == builder2._engine_key)
    assert jnp.all(builder._jitter_key == builder2._jitter_key)


def test_jitter_fns():
    con = DictInterface(lambda ms: -0.5 * ms["x"] ** 2 - 0.5 * ms["y"])
    ms = {"x": jnp.array(1), "y": jnp.array(-1)}

    num_chains = 2

    builder = EngineBuilder(seed=1, num_chains=num_chains)
    builder.set_model(con)
    builder.set_initial_values(ms, multiple_chains=False)
    builder.set_jitter_fns(
        {
            "x": (
                lambda key, cv: cv
                + tfd.Uniform(-1.0, 1.0).sample(sample_shape=cv.shape, seed=key)
            ),
            "y": (
                lambda key, cv: cv
                + tfd.Uniform(-1.0, 1.0).sample(sample_shape=cv.shape, seed=key)
            ),
        }
    )
    builder.add_kernel(gs.IWLSKernel(["x", "y"]))
    builder.set_duration(warmup_duration=200, posterior_duration=10, term_duration=10)
    engine = builder.build()

    assert not jnp.allclose(ms["x"], engine._model_states["x"][0])
    assert not jnp.allclose(ms["y"], engine._model_states["y"][0])
    assert not jnp.allclose(ms["x"], engine._model_states["x"][1])
    assert not jnp.allclose(ms["y"], engine._model_states["y"][1])

    assert not jnp.allclose(engine._model_states["x"][0], engine._model_states["x"][1])



"""
The model and the graph builder.
"""

from __future__ import annotations

import logging
import math
import re
from collections import Counter
from collections.abc import Iterable, Sequence
from copy import deepcopy
from types import MappingProxyType
from typing import IO, Any, Literal, TypeVar

import dill
import jax
import jax.numpy as jnp
import jax.random
import networkx as nx

from .nodes import Array, Calc, Dist, Group, Node, NodeState, Value, Var, VarValue
from .viz import plot_nodes, plot_vars

__all__ = ["GraphBuilder", "Model", "load_model", "save_model"]

logger = logging.getLogger(__name__)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Graph builder ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


NV = TypeVar("NV", Node, Var)


def _reduced_sum(*args: Array) -> Array:
    """Computes the sum after reducing arrays to scalars."""
    reduced = (arg.sum() if hasattr(arg, "sum") else arg for arg in args)
    return sum(reduced)


def _transform_back(var_transformed: Var) -> Calc:
    """
    Creates a :class:`.Calc` mapping a transformed parameter back to
    the original domain.
    """

    if var_transformed.dist_node is None:
        raise RuntimeError(
            f"{repr(var_transformed)} must have a transformed distribution"
        )

    transformed_distribution = var_transformed.dist_node.distribution

    def fn(at, *args, **kwargs):
        bijector = transformed_distribution(*args, **kwargs).bijector
        return bijector.inverse(at)

    inputs = var_transformed.dist_node.inputs
    kwinputs = var_transformed.dist_node.kwinputs

    return Calc(fn, var_transformed.value_node, *inputs, **kwinputs)  # type: ignore


class GraphBuilder:
    """
    A graph builder, used to set up a :class:`.Model`.

    Constructs a model containing all nodes and variables that were added to the graph
    builder and their recursive inputs.

    .. important::
        - In :meth:`.build_model` , the graph builder will automatically find all
          **inputs** to its nodes - and the inputs to these inputs
          (i.e. it finds inputs recursively).
        - The **outputs** of the nodes, however, are not added to the model
          automatically, so all **root nodes** need to be added explicitly.
        - Root nodes are nodes that are not inputs to any other node in the graph.
          The response in a regression model is an example of a root node.

    The standard workflow is to create the nodes and variables, add the root var to a
    graph builder, and construct a model from the graph builder. After the model has
    been constructed, some methods of the graph builder are not available anymore,
    because the graph is considered static.

    Parameters
    ----------
    to_float32
        Whether to convert the dtype of the values of the added nodes \
        from float64 to float32.

    See Also
    --------

    :class:`.Model` : The liesel model class, representing a static graph.
    :meth:`.GraphBuilder.add` : Method for adding variables and nodes to the
        GraphBuilder.
    :meth:`.GraphBuilder.build_model` : Method for building a model from the
        GraphBuilder.
    :meth:`.Var.transform` : Transforms a variable by adding a new transformed
        variable as an input. This is useful for variables that are constrained to a
        certain domain, e.g. positive values.

    Examples
    --------

    We start by creating some variables:

    >>> a = lsl.Var(1.0, name="a")
    >>> b = lsl.Var(2.0, name="b")
    >>> c = Var.new_calc(lambda x, y: x + y, a, b, name="c")

    We now initialize a GraphBuilder and add the root node ``c`` to it:

    >>> gb = lsl.GraphBuilder()
    >>> gb.add(c)
    GraphBuilder(0 nodes, 1 vars)

    We are now ready to build the model:

    >>> model = gb.build_model()
    >>> model
    Model(9 nodes, 3 vars)

    Note that when :meth:`.build_model` is called, all :attr:`~.Var.weak` variables in
    the graph will be updated. So the value of ``c`` is now available:

    >>> c.value
    3.0

    The graph builder is now empty:

    >>> gb.vars
    []
    """

    def __init__(self, to_float32: bool = True):
        self.nodes: list[Node] = []
        """The nodes that were explicitly added to the graph."""

        self.vars: list[Var] = []
        """The variables that were explicitly added to the graph."""

        self._log_lik_node: Node | None = None
        self._log_prior_node: Node | None = None
        self._log_prob_node: Node | None = None
        self.to_float32 = to_float32

    def _add_model_log_lik_node(self) -> GraphBuilder:
        """Adds the model log-likelihood node with the name ``_model_log_lik``."""

        if self.log_lik_node:
            self.add(
                Calc(
                    lambda x: x,
                    self.log_lik_node,
                    _name="_model_log_lik",
                    _update_on_init=False,
                )
            )
            return self

        _, _vars = self._all_nodes_and_vars()
        inputs = (v.dist_node for v in _vars if v.has_dist and v.observed)
        node = Calc(
            _reduced_sum, *inputs, _name="_model_log_lik", _update_on_init=False
        )
        self.add(node)
        return self

    def _add_model_log_prior_node(self) -> GraphBuilder:
        """Adds the model log-prior node with the name ``_model_log_prior``."""

        if self.log_prior_node:
            self.add(
                Calc(
                    lambda x: x,
                    self.log_prior_node,
                    _name="_model_log_prior",
                    _update_on_init=False,
                )
            )
            return self

        _, _vars = self._all_nodes_and_vars()
        inputs = (v.dist_node for v in _vars if v.has_dist and v.parameter)
        node = Calc(
            _reduced_sum, *inputs, _name="_model_log_prior", _update_on_init=False
        )
        self.add(node)
        return self

    def _add_model_log_prob_node(self) -> GraphBuilder:
        """Adds the model log-probability node with the name ``_model_log_prob``."""

        if self.log_prob_node:
            self.add(
                Calc(
                    lambda x: x,
                    self.log_prob_node,
                    _name="_model_log_prob",
                    _update_on_init=False,
                )
            )
            return self

        nodes, _ = self._all_nodes_and_vars()
        inputs = (n for n in nodes if isinstance(n, Dist))
        node = Calc(
            _reduced_sum, *inputs, _name="_model_log_prob", _update_on_init=False
        )
        self.add(node)
        return self

    def _add_model_seed_nodes(self) -> GraphBuilder:
        """Adds the model seed nodes with the names ``_model_*_seed``."""
        nodes, _ = self._all_nodes_and_vars()

        for node in nodes:
            if node.needs_seed:
                seed = Value(jax.random.PRNGKey(0), _name=f"_model_{node.name}_seed")
                node.set_inputs(*node.inputs, **{"seed": seed} | node.kwinputs)

        return self

    def _all_nodes_and_vars(self) -> tuple[list[Node], list[Var]]:
        """
        Returns all nodes and variables that were explicitly or implicitly
        (as recursive inputs) added to the graph.
        """
        nodes = self.nodes.copy()
        nodes.extend(node for var in self.vars for node in var.nodes)

        nodes = list(dict.fromkeys(nodes))

        if self.log_lik_node:
            nodes.append(self.log_lik_node)

        if self.log_prior_node:
            nodes.append(self.log_prior_node)

        if self.log_prob_node:
            nodes.append(self.log_prob_node)

        all_nodes: list[Node] = []
        all_vars: list[Var] = []

        while nodes:
            node = nodes.pop()

            if node in all_nodes:
                continue

            nodes.extend(node.all_input_nodes())
            all_nodes.append(node)

            if node.var:
                if node.var in all_vars:
                    continue

                nodes.extend(node.var.nodes)
                all_vars.append(node.var)

        return all_nodes, all_vars

    @staticmethod
    def _do_set_missing_names(nodes_or_vars: Iterable[NV], prefix: str) -> list[str]:
        """Sets the missing names for the given nodes or variables."""
        other = [nv.name for nv in nodes_or_vars if nv.name]
        counter = -1

        automatically_set_names = []

        for nv in nodes_or_vars:
            if not nv.name:
                name = f"{prefix}{(counter := counter + 1)}"

                while name in other:
                    name = f"{prefix}{(counter := counter + 1)}"

                nv.name = name
                other.append(name)
                automatically_set_names.append(name)

        return automatically_set_names

    def _set_missing_names(self) -> dict[str, list[str]]:
        """Sets the missing node and variable names."""
        nodes, _vars = self._all_nodes_and_vars()
        auto_var_names = self._do_set_missing_names(_vars, prefix="v")
        auto_node_names = self._do_set_missing_names(nodes, prefix="n")
        return {"vars": auto_var_names, "nodes": auto_node_names}

    def add(
        self, *args: Node | Var | GraphBuilder, to_float32: bool | None = None
    ) -> GraphBuilder:
        """
        Adds nodes, variables or other graph builders to the graph.

        Parameters
        ----------
        *args
            The nodes, variables or graph builders to add to the graph. Note that \
            the GraphBuilder will find input nodes recursively for all nodes and \
            variables that are added to it, so you only need to add root nodes.
        to_float32
            Whether to convert the dtype of the values of the added nodes \
            from float64 to float32. If ``None`` (default), the GraphBuilder's \
            attribute ``GraphBuilder.to_float32``, which is set during initialization \
            will be used instead.

        See Also
        --------
        :meth:`.GraphBuilder.build_model` : Method for building a model from the \
            GraphBuilder.
        :meth:`.Var.transform` : Transforms a variable by adding a new
            transformed variable as an input.

        Examples
        --------

        We start by creating some variables:

        >>> a = lsl.Var(1.0, name="a")
        >>> b = lsl.Var(2.0, name="b")
        >>> c = Var.new_calc(lambda x, y: x + y, a, b, name="c")

        We now initialize a GraphBuilder and add the root node ``c`` to it:

        >>> gb = lsl.GraphBuilder()
        >>> gb.add(c)
        GraphBuilder(0 nodes, 1 vars)

        We are now ready to build the model:

        >>> model = gb.build_model()
        >>> model
        Model(9 nodes, 3 vars)
        """

        if to_float32 is None:
            to_float32 = self.to_float32

        for arg in args:
            if isinstance(arg, Node):
                self.nodes.append(arg)
            elif isinstance(arg, Var):
                self.vars.append(arg)
            elif isinstance(arg, GraphBuilder):
                self.nodes.extend(arg.nodes)
                self.vars.extend(arg.vars)
            else:
                raise RuntimeError(f"Cannot add {type(arg).__name__} to graph builder")

        if to_float32:
            self.convert_dtype("float64", "float32")

        return self

    def add_groups(
        self, *groups: Group, to_float32: bool | None = None
    ) -> GraphBuilder:
        """
        Adds groups to the graph.

        Parameters
        ----------
        *groups
            The groups to add to the graph.
        to_float32
            Whether to convert the dtype of the values of the added nodes \
            from float64 to float32. If ``None`` (default), the GraphBuilder's \
            attribute ``GraphBuilder.to_float32``, which is set during initialization \
            will be used instead.

        Returns
        -------
        The graph builder.
        """

        if to_float32 is None:
            to_float32 = self.to_float32

        for group in groups:
            old = self.groups()

            if group.name in old and group is not old[group.name]:
                raise RuntimeError(
                    f"Group with name {repr(group.name)} already exists "
                    "in graph builder"
                )

            self.add(*group.nodes_and_vars.values())

        if to_float32:
            self.convert_dtype("float64", "float32")

        return self

    def build_model(self, copy: bool = False) -> Model:
        """
        Builds a model from the graph.

        Constructs a model containing all nodes and variables that were added to the
        graph builder and their recursive inputs. The outputs of the nodes are not added
        to the model automatically, so the root nodes always need to be added
        explicitly.

        The standard workflow is to create the nodes and variables, add them to a graph
        builder, and construct a model from the graph builder. After the model has been
        constructed, some methods of the graph builder are not available anymore.

        Parameters
        ----------
        copy
            Whether the nodes and variables should be copied when building the model.

        Returns
        -------
        The liesel model, which is a static graph built from the GraphBuilder.

        Notes
        -----
        If this method is called with the argument ``copy=False``, all nodes and
        variables are removed from the graph builder, because most methods of the graph
        builder do not work with nodes that are part of a model.

        Examples
        --------

        We start by creating some variables:

        >>> a = lsl.Var(1.0, name="a")
        >>> b = lsl.Var(2.0, name="b")
        >>> c = Var.new_calc(lambda x, y: x + y, a, b, name="c")

        We now initialize a GraphBuilder and add the root node ``c`` to it:

        >>> gb = lsl.GraphBuilder()
        >>> gb.add(c)
        GraphBuilder(0 nodes, 1 vars)

        We are now ready to build the model:

        >>> model = gb.build_model()
        >>> model
        Model(9 nodes, 3 vars)

        Note that when :meth:`.build_model` is called, all :attr:`~.Var.weak` variables
        in the graph will be updated. So the value of ``c`` is now available:

        >>> c.value
        3.0

        The graph builder is now empty:

        >>> gb.vars
        []
        """
        nodes, _vars = self._all_nodes_and_vars()

        if not nodes:
            logger.warning("No nodes in graph builder, building an empty model")

        for node in nodes:
            if node.name.startswith("_model"):
                raise RuntimeError(f"{repr(node)} has reserved name '_model*'")

        gb = self.copy()

        nodes, _vars = gb._all_nodes_and_vars()

        for var in _vars:
            if var.auto_transform:
                if var.dist_node is None:
                    raise RuntimeError(
                        f"Auto-transform of {var} failed, because it has no"
                        " distribution, which means no default bijector can be found."
                    )
                tname = f"{var.name}_transformed"
                if tname in nodes or tname in _vars:
                    raise RuntimeError(
                        f"Auto-transform of {var} failed, because a variable of the "
                        f"name {tname} is already present in {gb}."
                    )
                var.transform(bijector=None)

        gb._set_missing_names()
        gb._add_model_log_lik_node()
        gb._add_model_log_prior_node()
        gb._add_model_log_prob_node()
        gb._add_model_seed_nodes()

        nodes, _vars = gb._all_nodes_and_vars()
        nodes_and_vars = nodes + _vars

        model = Model(nodes_and_vars, grow=False, copy=copy)

        if not copy:
            self.nodes.clear()
            self.vars.clear()

            self._log_lik_node = None
            self._log_prior_node = None
            self._log_prob_node = None

        return model

    def convert_dtype(
        self, from_dtype: str | jax.numpy.dtype, to_dtype: str | jax.numpy.dtype
    ) -> GraphBuilder:
        """
        Tries to convert the node values in the graph to the specified data type.

        Works for nodes whose value is an array or pytree_. Nodes whose value is of
        another type are silently ignored.

        .. _pytree: https://jax.readthedocs.io/en/latest/pytrees.html

        Parameters
        ----------
        from_dtype
            The data type to convert from.
        to_dtype
            The data type to convert to.

        Returns
        -------
        The graph builder.


        """
        nodes, _ = self._all_nodes_and_vars()

        class ConversionWrapper:
            def __init__(self, value):
                self.value = value
                self.converted = False

                try:
                    if value.dtype == from_dtype:
                        self.value = value.astype(to_dtype)
                        self.converted = True
                except AttributeError:
                    pass

        for node in nodes:
            try:
                wrappers = jax.tree.map(ConversionWrapper, node.value)

                value = jax.tree.map(lambda x: x.value, wrappers)
                node.value = value  # type: ignore # data node

                converted = jax.tree.map(lambda x: x.converted, wrappers)

                if any(jax.tree_util.tree_flatten(converted)[0]):
                    logger.info(f"Converted dtype of {repr(node)}.value")
            except AttributeError:
                pass

        return self

    def copy(self) -> GraphBuilder:
        """Returns a shallow copy of the graph builder."""
        gb = GraphBuilder(to_float32=self.to_float32)
        gb.nodes = self.nodes.copy()
        gb.vars = self.vars.copy()

        gb.log_lik_node = self.log_lik_node
        gb.log_prior_node = self.log_prior_node
        gb.log_prob_node = self.log_prob_node

        return gb

    def count_node_names(self) -> dict[str, int]:
        """Counts the number of times each node name occurs in the graph."""
        nodes, _ = self._all_nodes_and_vars()
        counter = Counter(node.name for node in nodes if node.name)
        return dict(counter.most_common())

    def count_var_names(self) -> dict[str, int]:
        """Counts the number of times each variable name occurs in the graph."""
        _, _vars = self._all_nodes_and_vars()
        counter = Counter(var.name for var in _vars if var.name)
        return dict(counter.most_common())

    def groups(self) -> dict[str, Group]:
        """Collects the groups from all nodes and variables."""
        nodes, _vars = self._all_nodes_and_vars()
        g1 = {g.name: g for n in nodes for g in n.groups.values()}
        g2 = {g.name: g for v in _vars for g in v.groups.values()}
        return g1 | g2

    @property
    def log_lik_node(self) -> Node | None:
        """User-defined log-likelihood node, if there is one."""
        return self._log_lik_node

    @log_lik_node.setter
    def log_lik_node(self, log_lik_node: Node | None):
        if log_lik_node and not isinstance(log_lik_node, Node):
            raise RuntimeError("The log-likelihood node must be a node, not var")

        self._log_lik_node = log_lik_node

    @property
    def log_prior_node(self) -> Node | None:
        """User-defined log-prior node, if there is one."""
        return self._log_prior_node

    @log_prior_node.setter
    def log_prior_node(self, log_prior_node: Node | None):
        if log_prior_node and not isinstance(log_prior_node, Node):
            raise RuntimeError("The log-prior node must be a node, not var")

        self._log_prior_node = log_prior_node

    @property
    def log_prob_node(self) -> Node | None:
        """User-defined log-probability node, if there is one."""
        return self._log_prob_node

    @log_prob_node.setter
    def log_prob_node(self, log_prob_node: Node | None):
        if log_prob_node and not isinstance(log_prob_node, Node):
            raise RuntimeError("The log-probability node must be a node, not var")

        self._log_prob_node = log_prob_node

    def plot_nodes(self) -> GraphBuilder:
        """
        Plots all nodes in the graph.

        See Also
        --------
        :meth:`.viz.plot_nodes` : The function used to plot the nodes.

        """
        nodes, _vars = self._all_nodes_and_vars()
        nodes_and_vars = nodes + _vars

        self._set_missing_names()
        model = Model(nodes_and_vars, grow=False)
        plot_nodes(model)

        model.pop_nodes_and_vars()

        return self

    def plot_vars(self) -> GraphBuilder:
        """
        Plots all variables in the graph.

        Returns
        -------
        The graph builder.

        See Also
        --------
        :meth:`.viz.plot_vars` : The function used to plot the variables.
        """
        nodes, _vars = self._all_nodes_and_vars()
        nodes_and_vars = nodes + _vars

        self._set_missing_names()
        model = Model(nodes_and_vars, grow=False)
        plot_vars(model)

        model.pop_nodes_and_vars()

        return self

    def rename(self, pattern: str, replacement: str) -> GraphBuilder:
        """Renames all nodes and variables in the graph."""
        self.rename_nodes(pattern, replacement)
        self.rename_vars(pattern, replacement)
        return self

    def rename_nodes(self, pattern: str, replacement: str) -> GraphBuilder:
        """Renames all nodes in the graph."""
        nodes, _ = self._all_nodes_and_vars()

        for node in nodes:
            if node.name:
                node.name = re.sub(pattern, replacement, node.name)

        return self

    def rename_vars(self, pattern: str, replacement: str) -> GraphBuilder:
        """Renames all variables in the graph."""
        _, _vars = self._all_nodes_and_vars()

        for var in _vars:
            if var.name:
                var.name = re.sub(pattern, replacement, var.name)

        return self

    def replace_node(self, old: Node, new: Node) -> GraphBuilder:
        """Replaces the ``old`` with the ``new`` node."""
        self.nodes = [new if x is old else x for x in self.nodes]
        nodes, _ = self._all_nodes_and_vars()

        for node in nodes:
            inputs = [new if x is old else x for x in node.inputs]
            kwinputs = {k: new if v is old else v for k, v in node.kwinputs.items()}
            node.set_inputs(*inputs, **kwinputs)

        return self

    def replace_var(self, old: Var, new: Var) -> GraphBuilder:
        """Replaces the ``old`` with the ``new`` variable."""
        self.vars = [new if x is old else x for x in self.vars]
        self.replace_node(old.var_value_node, new.var_value_node)
        self.replace_node(old.value_node, new.value_node)

        if old.dist_node:
            if not new.dist_node:
                raise RuntimeError(
                    f"Cannot replace {repr(old)} with distribution "
                    f"with {repr(new)} without distribution"
                )

            self.replace_node(old.dist_node, new.dist_node)

        return self

    def update(self) -> GraphBuilder:
        """
        Updates all nodes in the graph.

        Returns
        -------
        The graph builder.
        """
        nodes, _vars = self._all_nodes_and_vars()
        nodes_and_vars = nodes + _vars

        self._set_missing_names()
        model = Model(nodes_and_vars, grow=False)
        model.pop_nodes_and_vars()

        return self

    def __repr__(self) -> str:
        brackets = f"({len(self.nodes)} nodes, {len(self.vars)} vars)"
        return type(self).__name__ + brackets


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Model ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class Model:
    """
    A model with a static graph.

    .. tip::
        If you have an existing model and want to make changes to it, you can use the
        :meth:`.Model.pop_nodes_and_vars` method to release the nodes and variables
        of the model. You can then make changes to them, for example i.e. changing the
        distribution of a variable or the inputs of a calculation. Afterwards, you
        initialize a *new* model with your changed variables.
        If you simply want to change the value of a variable, it is not necessary to
        call :meth:`~.Model.pop_nodes_and_vars`, you can simply override the
        :attr:`.Var.value` attribute. Remeber to call :meth:`.Model.update` afterwards.

    Parameters
    ----------
    nodes_and_vars
        The nodes and variables to include in the model.
    grow
        Whether a :class:`.GraphBuilder` should be used to grow the model (finding \
        the recursive inputs of the nodes and variables), and to add the model nodes.
    copy
        Whether the nodes and variables should be copied upon initialization.
    to_float32
        Whether to convert the dtype of the values of the added nodes \
        from float64 to float32. Only takes effect if ``grow=True``.

    See Also
    --------
    .Var.new_obs : Initializes a strong variable that holds observed data.
    .Var.new_param : Initializes a strong variable that acts as a model parameter.
    .Var.new_calc :
        Initializes a weak variable that is a function of other variables.
    .Var.new_value : Initializes a strong variable without a distribution.
    :class:`.GraphBuilder` :
        A graph builder, which can be used to set up and manipulate a model if you need
        more control.

    Examples
    --------

    Here, we set up a basic model based on three variables:

    >>> a = lsl.Var.new_value(1.0, name="a")
    >>> b = lsl.Var.new_value(2.0, name="b")
    >>> c = lsl.Var.new_calc(lambda x, y: x + y, a, b, name="c")

    We now build a model:

    >>> model = lsl.Model([c])
    >>> model
    Model(9 nodes, 3 vars)

    """

    def __init__(
        self,
        nodes_and_vars: Iterable[Node | Var],
        grow: bool = True,
        copy: bool = False,
        to_float32: bool = True,
    ):
        self._to_float32 = to_float32
        if grow:
            model = (
                GraphBuilder(to_float32=to_float32).add(*nodes_and_vars).build_model()
            )
            nodes_and_vars = [*model.nodes.values(), *model.vars.values()]
            model.pop_nodes_and_vars()

        nodes = [nv for nv in nodes_and_vars if isinstance(nv, Node)]
        nodes = list(dict.fromkeys(nodes).keys())
        counts = Counter(n.name for n in nodes)
        dups = [k for k, v in counts.items() if v > 1]

        if dups:
            raise RuntimeError(f"Duplicate node names: {', '.join(dups)}")

        _vars = [nv for nv in nodes_and_vars if isinstance(nv, Var)]
        _vars = list(dict.fromkeys(_vars).keys())
        counts = Counter(v.name for v in _vars)
        dups = [k for k, v in counts.items() if v > 1]

        if dups:
            raise RuntimeError(f"Duplicate variable names: {', '.join(dups)}")

        groups = [g for nv in nodes_and_vars for g in nv.groups.values()]
        groups = list(dict.fromkeys(groups).keys())
        counts = Counter(g.name for g in groups)
        dups = [k for k, v in counts.items() if v > 1]

        if dups:
            raise RuntimeError(f"Duplicate group names: {', '.join(dups)}")

        self._nodes = {n.name: n for n in nodes}
        self._vars = {v.name: v for v in _vars}

        if copy:
            self._nodes, self._vars = deepcopy((self._nodes, self._vars))

        for node in self._nodes.values():
            node._clear_outputs()
            node._set_model(self)

        for node in self._nodes.values():
            for _input in node.all_input_nodes():
                _input._add_output(node)

        self._node_graph = self._build_node_graph(self._nodes.values())
        self._var_graph = self._build_var_graph(self._vars.values())

        self._sorted_nodes = list(nx.topological_sort(self._node_graph))

        self._simulation_graph = self._build_simulation_graph(self._nodes.values())
        self._simulation_nodes = list(nx.topological_sort(self._simulation_graph))

        self._auto_update = True
        self._seed_nodes = []

        for node in self._sorted_nodes:
            if node.name.startswith("_model_") and node.name.endswith("_seed"):
                self._seed_nodes.append(node)

            node.update()

    @staticmethod
    def _build_node_graph(nodes: Iterable[Node]) -> nx.DiGraph:
        """Builds the directed graph of the model nodes."""
        edges: list[tuple[Node, Node]] = []

        for node in nodes:
            edges.extend((_input, node) for _input in node.all_input_nodes())

        graph = nx.DiGraph(edges)
        graph.add_nodes_from(nodes)
        return graph

    @staticmethod
    def _build_simulation_graph(nodes: Iterable[Node]) -> nx.DiGraph:
        """Builds the simulation graph of the model nodes."""
        edges: list[tuple[Node, Node]] = []

        for node in nodes:
            for _input in node.all_input_nodes():
                if isinstance(node, Dist) and _input is node.at:
                    edges.append((node, _input))
                else:
                    edges.append((_input, node))

        graph = nx.DiGraph(edges)
        graph.add_nodes_from(nodes)
        return graph

    @staticmethod
    def _build_var_graph(_vars: Iterable[Var]) -> nx.DiGraph:
        """Builds the directed graph of the model variables."""
        edges: list[tuple[Var, Var]] = []

        for var in _vars:
            edges.extend((_input, var) for _input in var.all_input_vars())

        graph = nx.DiGraph(edges)
        graph.add_nodes_from(_vars)
        return graph

    def _copy_computational_model(self) -> Model:
        """Returns a deep copy of the model with all node states cleared."""
        backup = self.state

        for node in self._nodes.values():
            node.clear_state()

        empty = deepcopy(self)
        self.state = backup

        return empty

    def _recursive_inputs(self, name: str) -> list[Node]:
        """Returns the recursive inputs of a model node."""
        nodes = [self._nodes[name]]
        visited = []

        while nodes:
            node = nodes.pop()

            if node in visited:
                continue

            nodes.extend(node.all_input_nodes())
            visited.append(node)

        return visited

    @property
    def auto_update(self) -> bool:
        """
        Whether to update the model automatically if the value of a node is modified.

        The auto-update can be disabled to improve the performance if multiple model
        parameters are updated at once.
        """
        return self._auto_update

    @auto_update.setter
    def auto_update(self, auto_update: bool):
        self._auto_update = auto_update

    def groups(self) -> dict[str, Group]:
        """Collects the groups from all nodes and variables."""
        g1 = {g.name: g for n in self._nodes.values() for g in n.groups.values()}
        g2 = {g.name: g for v in self._vars.values() for g in v.groups.values()}
        return g1 | g2

    def copy_nodes_and_vars(self) -> tuple[dict[str, Node], dict[str, Var]]:
        """Returns an unfrozen deep copy of the model nodes and variables."""
        nodes, _vars = deepcopy((self._nodes, self._vars))

        for node in nodes.values():
            node._unset_model()

        nodes = {nm: nd for nm, nd in nodes.items() if not nm.startswith("_model")}

        return nodes, _vars

    def node_parental_subgraph(self, *of: Node) -> nx.DiGraph:
        """
        Returns a subgraph that consists of the input nodes and their parent nodes.
        """
        nodes_to_include = set()
        for node in of:
            nodes_to_include.update(nx.ancestors(self.node_graph, node))
            nodes_to_include.add(node)
        subgraph = self.node_graph.subgraph(nodes_to_include)
        return subgraph

    def var_parental_subgraph(self, *of: Var) -> nx.DiGraph:
        """
        Returns a subgraph that consists of the input variables and their parent
        variables.
        """
        nodes_to_include = set()
        for node in of:
            nodes_to_include.update(nx.ancestors(self.var_graph, node))
            nodes_to_include.add(node)
        subgraph = self.var_graph.subgraph(nodes_to_include)
        return subgraph

    def parental_submodel(self, *of: Var | Node) -> Model:
        """
        Returns a new model that consists only of the given variables and nodes and \
        their parent variables and nodes. The new model contains copies of these \
        variables and nodes.
        """
        nodes_to_include = set()

        for node in of:
            if isinstance(node, Var):
                nodes_to_include.update(nx.ancestors(self.var_graph, node))
            else:
                nodes_to_include.update(nx.ancestors(self.node_graph, node))
            nodes_to_include.add(node)

        copy_of_nodes_to_include = deepcopy(nodes_to_include)

        for node in copy_of_nodes_to_include:
            if hasattr(node, "_unset_model"):
                node._unset_model()

        return Model(list(copy_of_nodes_to_include), to_float32=self._to_float32)

    @property
    def log_lik(self) -> Array:
        """
        The log-likelihood of the model.

        Defined as the sum of the log-probabilities of all observed variables
        with a probability distribution.
        """
        return self._nodes["_model_log_lik"].value

    @property
    def log_prior(self) -> Array:
        """
        The log-prior of the model.

        Defined as the sum of the log-probabilities of all parameter variables
        with a probability distribution.
        """
        return self._nodes["_model_log_prior"].value

    @property
    def log_prob(self) -> Array:
        """
        The (unnormalized) log-probability / log-posterior of the model.

        Defined as the sum of all distribution nodes.
        """
        return self._nodes["_model_log_prob"].value

    @property
    def node_graph(self) -> nx.DiGraph:
        """The directed graph of the model nodes."""
        return self._node_graph

    @property
    def nodes(self) -> MappingProxyType[str, Node]:
        """A mapping of the model nodes with their names as keys."""
        return MappingProxyType(self._nodes)

    def pop_nodes_and_vars(self) -> tuple[dict[str, Node], dict[str, Var]]:
        """
        Pops the nodes and variables out of this model.

        All nodes and variables are unfrozen and their reference to this model
        is removed. This model becomes invalid and cannot be used anymore.
        """
        nodes = self._nodes.copy()
        _vars = self._vars.copy()

        for node in nodes.values():
            node._unset_model()

        nodes = {nm: nd for nm, nd in nodes.items() if not nm.startswith("_model")}

        # clear the model
        self._nodes.clear()
        self._vars.clear()
        self._node_graph.clear()
        self._var_graph.clear()
        self._sorted_nodes.clear()
        self._seed_nodes.clear()

        return nodes, _vars

    def set_seed(self, seed: jax.Array) -> Model:
        """
        Splits and sets the seed / PRNG key.

        Parameters
        ----------
        seed
            The seed is split and distributed to the seed nodes of the model.
            Must be a jax RNG key array that satisfies
            ``jnp.issubdtype(key.dtype, jax.dtypes.prng_key)``.
            See :mod:`jax.random` and
            https://docs.jax.dev/en/latest/jep/9263-typed-keys.html for more details.
        """
        seeds = jax.random.split(seed, len(self._seed_nodes))

        for node, seed in zip(self._seed_nodes, seeds):
            node.value = seed  # type: ignore  # data node

        return self

    def simulate(self, seed: jax.Array, skip: Iterable[str] = ()) -> Model:
        """
        Updates the model state simulating from the probability distributions in the
        model using a provided random seed, optionally skipping specified nodes.

        Parameters
        ----------
        seed
            The seed is split and distributed to the seed nodes of the model. \
            Must be a jax RNG key array that satisfies \
            ``jnp.issubdtype(key.dtype, jax.dtypes.prng_key)``. \
            See :mod:`jax.random` and \
            https://docs.jax.dev/en/latest/jep/9263-typed-keys.html for more details.
        skip
            The names of the nodes or variables to be excluded from the simulation. \
            By default, no nodes or variables are skipped.

        Returns
        -------
        The model instance itself after updating its state with the simulated values.

        Raises
        ------
        AttributeError
            If the value of the :attr:`.Dist.at` node of a distribution node cannot be
            set.

        Notes
        -----
        The simulation is based on the shapes of the current values of the
        :attr:`.Dist.at` nodes of the distribution nodes. If the :attr:`.Dist.at` node
        of a distribution node is a :Class:`.VarValue` node, the value of its input is
        updated.
        """
        dists = [
            node
            for node in self._simulation_nodes
            if isinstance(node, Dist)
            and node.at is not None
            and node.name not in skip
            and node.at.name not in skip
            and (node.var is not None and node.var.name not in skip)
        ]

        seeds = jax.random.split(seed, len(dists))

        for dist, seed in zip(dists, seeds):
            tfp_dist = dist.init_dist()

            event_shape = tfp_dist.event_shape
            batch_shape = tfp_dist.batch_shape
            value_shape = jnp.asarray(dist.at.value).shape  # type: ignore
            sample_index = len(value_shape) - len(batch_shape) - len(event_shape)
            sample_shape = value_shape[:sample_index]

            value = tfp_dist.sample(sample_shape, seed)

            if isinstance(dist.at, VarValue):
                try:
                    dist.at.inputs[0].value = value  # type: ignore
                except AttributeError:
                    raise AttributeError(f"Cannot set value of {dist.at.inputs[0]}")
            else:
                try:
                    dist.at.value = value  # type: ignore
                except AttributeError:
                    raise AttributeError(f"Cannot set value of {dist.at}")

        return self

    def sample(
        self,
        shape: Sequence[int],
        seed: jax.Array,
        posterior_samples: dict[str, Array] | None = None,
        fixed: Sequence[str] = (),
        newdata: dict[str, Array] | None = None,
        dists: dict[str, Dist] | None = None,
    ) -> dict[str, Array]:
        """
        Draws samples from the model.

        Parameters
        ----------
        shape
            Sample shape.
        seed
            The seed is split and distributed to the seed nodes of the model. \
            Must be a jax RNG key array that satisfies \
            ``jnp.issubdtype(key.dtype, jax.dtypes.prng_key)``. \
            See :mod:`jax.random` and \
            https://docs.jax.dev/en/latest/jep/9263-typed-keys.html for more details.
        posterior_samples
            Dictionary of samples at which to evaluate predictions. All values of the \
            dictionary are assumed to have two leading dimensions corresponding to \
            ``(nchains, niteration)``.
        fixed
            The names of the nodes or variables to be excluded from the simulation. \
            By default, no nodes or variables are skipped.
        newdata
            Dictionary of new data at which to produce samples. The keys should \
            correspond to variable or node names in the model whose values should be \
            set to the given values before sampling. If ``None`` \
            (default), the current variable values are used.
        dists
            Can be used to provide a dictionary of variable names and :class:`.Dist` \
            instances to use in sampling. If ``None`` (default), samples are drawn for \
            each variable using their :attr:`.Var.dist_node`.

        Notes
        -----
        When compiling this function with ``jax.jit``, the arguments ``shape``,
        ``fixed``, and ``dists`` must be static.

        Returns
        -------
        A dictionary of variable and node names and their sampled values. Includes
        only sampled variables.
        """
        # Pre-processing
        # ------------------------------------------------------------------------------
        posterior_samples = posterior_samples if posterior_samples is not None else {}
        state_for_sampling = (
            self.update_state(newdata) if newdata is not None else self.state
        )
        dists = dists if dists is not None else {}

        # Input validation
        # ------------------------------------------------------------------------------

        # validate values in dists
        for var_name in dists:
            vars_ = self.vars
            if var_name not in vars_:
                raise ValueError(f"No variable with name '{var_name}' found.")

            if vars_[var_name].weak:
                raise ValueError(f"Variable '{var_name}' is weak, cannot sample.")

        # validate consistency of 'fixed' and 'posterior_samples' arguments
        for name in fixed:
            if name in posterior_samples:
                raise ValueError(
                    f"Inconsistency: {name=} listed in 'fixed', but samples are"
                    " provided in 'posterior_samples'."
                )

        # Collect sampling information
        # ------------------------------------------------------------------------------

        # collect relevant distribution nodes in model
        dists_list = [
            node
            for node in self._simulation_nodes
            if isinstance(node, Dist)
            and node.at is not None
            and node.name not in fixed
            and node.at.name not in fixed
            and (node.var is not None and node.var.name not in fixed)
        ]

        # collect information for sampling by processing dist nodes
        sampling_specs = {}
        for i, dist in enumerate(dists_list):
            tfp_dist = dist.init_dist()

            event_shape = tfp_dist.event_shape
            batch_shape = tfp_dist.batch_shape
            value_shape = jnp.asarray(dist.at.value).shape  # type: ignore
            sample_index = len(value_shape) - len(batch_shape) - len(event_shape)
            sample_shape = value_shape[:sample_index]

            if isinstance(dist.at, VarValue):
                var_name = dist.at.var.name  # type: ignore
                value_var = dist.at.inputs[0]
            else:
                var_name = dist.at.name  # type: ignore
                value_var = dist.at  # type: ignore

            if var_name not in posterior_samples:
                # pulls manually defined distribution from dists dict, returns current
                # dist otherwise
                dist = dists.get(var_name, dist)

                sampling_specs[var_name] = {
                    "shape": sample_shape,
                    "dist": dist,
                    "i": i,
                    "value_var": value_var,
                }

        # add information for custom dists for variables that are not yet covered.
        for var_name, dist in dists.items():
            if var_name in sampling_specs:
                # in this case, the variable has already been added to sampling specs,
                # and it is also already using the custom dist
                continue

            i += 1
            tfp_dist = dist.init_dist()
            event_shape = tfp_dist.event_shape
            batch_shape = tfp_dist.batch_shape
            sample_index = len(value_shape) - len(batch_shape) - len(event_shape)
            sample_shape = value_shape[:sample_index]
            value_shape = jnp.asarray(self.vars[var_name].value).shape  # type: ignore

            value_var = self.vars[var_name].value_node

            sampling_specs[var_name] = {
                "shape": sample_shape,
                "dist": dist,
                "i": i,
                "value_var": value_var,
            }

        # Shape handling
        # ------------------------------------------------------------------------------

        # set up shape of samples
        samples_shape = (
            next(iter(posterior_samples.values())).shape[:2]
            if posterior_samples
            else ()
        )
        nsamples = math.prod(
            shape
        )  # total number of samples to draw (pure python so jit works)

        # set up all seeds that will be needed
        seeds = jax.random.split(
            seed, (nsamples,) + samples_shape + (len(sampling_specs),)
        )

        def reshape(a):
            # brings samples into the desired shape based on input argument.
            # shape=(3,4)
            # nsamples=12
            # shape of drawn samples: (12,...)
            # reshaped to (3,4, ...)
            return jnp.reshape(a, shape=shape + a.shape[1:])

        # Workhorse function
        # ------------------------------------------------------------------------------

        def one_draw(position, seeds):
            # the position argument is for updating the state with posterior samples
            previous_state = self.state

            # update model state using the position (a single posterior sample, if any)
            # and the state_for_sampling, which includes the observed values from
            # newdata.
            self.state = self.update_state(position, state_for_sampling)

            # draw samples in order of the model graph
            sampled_position = {}
            for name, spec in sampling_specs.items():
                # initializes the distribution node using the current model state,
                # which may have been influenced by 'position', 'newdata', or sampled
                # values from variables higher up the model hierarchy
                tfp_dist = spec["dist"].init_dist()

                # draw the actual sample
                value = tfp_dist.sample(spec["shape"], seeds[spec["i"]])

                # save the sampled value
                sampled_position[name] = value

                # update the variable's value with the sampled value so that the
                # distributions of variables further down the model hierarchy will be
                # correctly initialized based on the sampled values higher up
                spec["value_var"].value = value

            # to avoid tracer leakage we prevent side effects to persists
            self.state = previous_state

            return sampled_position

        if not posterior_samples:
            draw_chains = jax.vmap(one_draw, in_axes=(None, 0), out_axes=0)
            # since we have no posterior samples, we use position={}
            drawn_samples = draw_chains({}, seeds)

            # return reshaped version of samples
            return jax.tree.map(reshape, drawn_samples)

        # this branch of the function continues only if posterior_samples is not None
        # -----------------------------------------------------------------------------
        draw_iter = jax.vmap(one_draw, in_axes=(0, 0), out_axes=0)
        draw_chains = jax.vmap(draw_iter, in_axes=(0, 0), out_axes=0)
        draw_samples = jax.vmap(draw_chains, in_axes=(None, 0), out_axes=0)

        # filter samples to include only samples that belong to the model
        vars_and_nodes = list(self.vars) + list(self.nodes)
        filtered_samples = {
            k: v for k, v in posterior_samples.items() if k in vars_and_nodes
        }

        try:
            drawn_samples = draw_samples(filtered_samples, seeds)
        except Exception as e:
            msg = (
                "Error during sampling. Make sure to check sample shapes! The values in"
                " 'posterior_samples' must have two leading batching dimensions."
            )

            try:
                error_to_raise = e.__class__(msg)
            except Exception:
                # fallback in case e has a custom error class that cannot simply
                # be instantiated with a message.
                error_to_raise = RuntimeError(msg)

            raise error_to_raise from e

        # return reshaped version of samples
        return jax.tree.map(reshape, drawn_samples)

    @property
    def state(self) -> dict[str, NodeState]:
        """The state of the model as a dict of node names and states."""
        return {name: node.state for name, node in self._nodes.items()}

    @state.setter
    def state(self, state: dict[str, NodeState]):
        for name, node_state in state.items():
            self._nodes[name].state = node_state

    def update(self, *names: str) -> Model:
        """
        Updates the target nodes and their recursive inputs if they are outdated.

        The update is performed in a topological order, restoring a consistent state
        of the model. This method is called automatically by the nodes if their value
        is modified (unless :attr:`.auto_update` is ``False``).

        Parameters
        ----------
        names
            The names of the target nodes to be updated.
        """

        if not names:
            for node in self._sorted_nodes:
                if node.outdated:
                    node.update()
        else:
            inputs = set().union(*(self._recursive_inputs(name) for name in names))

            for node in self._sorted_nodes:
                if node in inputs and node.outdated:
                    node.update()

        return self

    @property
    def var_graph(self) -> nx.DiGraph:
        """The directed graph of the model variables."""
        return self._var_graph

    @property
    def vars(self) -> MappingProxyType[str, Var]:
        """A mapping of the model variables with their names as keys."""
        return MappingProxyType(self._vars)

    @property
    def parameters(self) -> MappingProxyType[str, Var]:
        """A mapping of the model parameters with their names as keys."""
        params = {k: v for k, v in self._vars.items() if v.parameter}
        return MappingProxyType(params)

    @property
    def observed(self) -> MappingProxyType[str, Var]:
        """A mapping of the observed model variables with their names as keys."""
        observed = {k: v for k, v in self._vars.items() if v.observed}
        return MappingProxyType(observed)

    def __repr__(self) -> str:
        brackets = f"({len(self._nodes)} nodes, {len(self._vars)} vars)"
        return type(self).__name__ + brackets

    def plot_vars(
        self,
        show: bool = True,
        save_path: str | None | IO = None,
        width: int = 14,
        height: int = 10,
        prog: Literal[
            "dot", "circo", "fdp", "neato", "osage", "patchwork", "sfdp", "twopi"
        ] = "dot",
    ):
        """
        Plots the variables of this model.

        Wraps :func:`~.viz.plot_vars`.

        Parameters
        ----------
        show
            Whether to show the plot in a new window.
        save_path
            Path to save the plot. If not provided, the plot will not be saved.
        width
            Width of the plot in inches.
        height
            Height of the plot in inches.
        prog
            Layout parameter. Available layouts: circo, dot (the default), fdp, neato, \
            osage, patchwork, sfdp, twopi.

        See Also
        --------
        .Var.plot_vars : Plots the variables of the Liesel sub-model that terminates in
            this variable.
        .Var.plot_nodes : Plots the nodes of the Liesel sub-model that terminates in
            this variable.
        .Model.plot_vars : Plots the variables of a Liesel model.
        .Model.plot_nodes : Plots the nodes of a Liesel model.
        .viz.plot_vars : Plots the variables of a Liesel model.
        .viz.plot_nodes : Plots the nodes of a Liesel model.
        """
        return plot_vars(
            self,
            show=show,
            save_path=save_path,
            width=width,
            height=height,
            prog=prog,
        )

    def plot_nodes(
        self,
        show: bool = True,
        save_path: str | None | IO = None,
        width: int = 14,
        height: int = 10,
        prog: Literal[
            "dot", "circo", "fdp", "neato", "osage", "patchwork", "sfdp", "twopi"
        ] = "dot",
    ):
        """
        Plots the nodes of this model.

        Wraps :func:`~.viz.plot_nodes`.

        Parameters
        ----------
        show
            Whether to show the plot in a new window.
        save_path
            Path to save the plot. If not provided, the plot will not be saved.
        width
            Width of the plot in inches.
        height
            Height of the plot in inches.
        prog
            Layout parameter. Available layouts: circo, dot (the default), fdp, neato, \
            osage, patchwork, sfdp, twopi.

        See Also
        --------
        .Var.plot_vars : Plots the variables of the Liesel sub-model that terminates in
            this variable.
        .Var.plot_nodes : Plots the nodes of the Liesel sub-model that terminates in
            this variable.
        .Model.plot_vars : Plots the variables of a Liesel model.
        .Model.plot_nodes : Plots the nodes of a Liesel model.
        .viz.plot_vars : Plots the variables of a Liesel model.
        .viz.plot_nodes : Plots the nodes of a Liesel model.
        """
        return plot_nodes(
            self,
            show=show,
            save_path=save_path,
            width=width,
            height=height,
            prog=prog,
        )

    def extract_position(
        self,
        position_keys: Sequence[str],
        model_state: dict[str, NodeState] | None = None,
    ) -> dict[str, Array]:
        """
        Extracts a position from a model state.

        Parameters
        ----------
        position_keys
            An iterable of variable or node names.
        model_state
            A dictionary of node names and their corresponding :class:`.NodeState`. \
            If ``None`` (default), the model's current state is used.
        """
        model_state = model_state if model_state is not None else self.state
        position = {}

        for key in position_keys:
            try:
                position[key] = model_state[key].value
            except KeyError:
                node_key = self.vars[key].value_node.name
                position[key] = model_state[node_key].value

        return position

    def update_state(
        self,
        position: dict[str, Array],
        model_state: dict[str, NodeState] | None = None,
        inplace: bool = False,
    ) -> dict[str, NodeState]:
        """
        Updates and returns a model state given a position.

        Parameters
        ----------
        position
            A dictionary of variable or node names and values.
        model_state
            A dictionary of node names and their corresponding :class:`.NodeState`. \
            If ``None`` (default), the model's current state is used.
        inplace
            If ``False`` (default), a new model state is returned, while the current \
            model's state is left unchanged. If ``True``, the current model's state is \
            updated in place.

        Warnings
        --------
        The ``model_state`` must be up-to-date, i.e. it must *not* contain any outdated
        nodes. Updates can only be triggered through new variable or node values in the
        ``position``. If you supply a ``model_state`` with outdated nodes, these nodes
        and their outputs will not be updated.
        """
        model = self._copy_computational_model() if not inplace else self

        # sets all outdated flags in the model state to false
        # this is required to make the function jittable

        model.state = model_state if model_state is not None else self.state

        for node in model.nodes.values():
            node._outdated = False

        # temporarily disable auto_update to avoid shape incompatibilities
        # when updating variables sequentially with new shapes
        original_auto_update = model.auto_update
        model.auto_update = False

        try:
            for key, value in position.items():
                try:
                    model.nodes[key].value = value  # type: ignore  # data node
                except KeyError:
                    model.vars[key].value = value
        finally:
            # restore original auto_update setting
            model.auto_update = original_auto_update

        model.update()
        return model.state

    def predict(
        self,
        samples: dict[str, Array],
        predict: Sequence[str] | None = None,
        newdata: dict[str, Array] | None = None,
    ) -> dict[str, Array]:
        """
        Returns a dictionary of predictions.

        Parameters
        ----------
        samples
            Dictionary of samples at which to evaluate predictions. All values of the \
            dictionary are assumed to have two leading dimensions corresponding to \
            ``(nchains, niteration)``.
        predict
            Sequence of strings, which are the names of nodes or variables. \
            Predictions will be returned only for the nodes or variables inlcuded \
            here. If ``None`` (default), predictions will be returned for all \
            *variables* in the model (but not for nodes).
        newdata
            Dictionary of new data at which to evaluate predictions. The keys should \
            correspond to variable or node names in the model whose values should be \
            set to the given values before evaluating predictions. If ``None`` \
            (default), the current variable values are used.
        """

        predict_names = predict

        # extract nodes and vars for target nodes
        if predict_names is None:
            # use full model without copying
            submodel = self
            predict_names = list(self.vars)  # output only vars
        else:
            predict_nodes_: list[Var | Node] = []
            for name in predict_names:
                try:
                    predict_nodes_.append(self.vars[name])
                except KeyError:
                    predict_nodes_.append(self.nodes[name])

            # construct submodel for target nodes
            submodel = self.parental_submodel(*predict_nodes_)

        # update submodel with new data, if any were given
        newdata = newdata if newdata is not None else {}
        submodel.state = submodel.update_state(newdata)

        # filter samples to include only samples that belong to the submodel
        vars_and_nodes = list(submodel.vars) + list(submodel.nodes)
        filtered_samples = {k: v for k, v in samples.items() if k in vars_and_nodes}
        if not filtered_samples:
            raise ValueError(
                "No samples provided for the variables or nodes in the submodel."
                f"Nodes in submodel: {vars_and_nodes}"
            )

        # single prediction function
        def predict_one(samples):
            updated_state = submodel.update_state(
                samples, submodel.state, inplace=False
            )
            return submodel.extract_position(predict_names, updated_state)

        # map over iterations
        predict_iter = jax.vmap(predict_one, in_axes=0, out_axes=0)

        # map over chains
        predict_chains = jax.vmap(predict_iter, in_axes=0, out_axes=0)

        # apply function
        return predict_chains(filtered_samples)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Save and load models ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def save_model(model: Any, file: str | IO[bytes]) -> None:
    """
    Saves a model to a `dill <https://github.com/uqfoundation/dill>`_ file.

    Parameters
    ----------
    model
        The model to be saved.
    file
        The file handler or path to save the model to.
    """

    if isinstance(file, str):
        with open(file, "wb") as handle:
            dill.dump(model, handle)
    else:
        dill.dump(model, file)


def load_model(file: str | IO[bytes]) -> Any:
    """
    Loads a model from a `dill <https://github.com/uqfoundation/dill>`_ file.

    Parameters
    ----------
    file
        The file handler or path to load the model from.
    """

    if isinstance(file, str):
        with open(file, "rb") as handle:
            model = dill.load(handle)
    else:
        model = dill.load(file)

    return model


class TemporaryModel:
    """
    A contextmanager for creating a temporary model.

    Unnamed variables and nodes will be named temporarily. Their names will be reset
    when the context closes.

    Example use::

        with TemporaryModel(variable) as model:
            print(model.log_prob)

    Parameters
    ----------
    vars_and_nodes
        Variable-length collection of variables and nodes that should be used to build \
        the temporary model.
    verbose
        If ``verbose=True``, the detailed information about temporarily namend \
        variables will be logged. If ``verbose=False``,  this is reduced to a short
        note.
    silent
        If ``silent=True``, all logging will be suppressed.
    """

    def __init__(
        self,
        *vars_and_nodes,
        verbose: bool = False,
        silent: bool = False,
        to_float32: bool = True,
    ):
        self.vars_and_nodes = vars_and_nodes
        self.verbose = verbose
        self.silent = silent
        self.to_float32 = to_float32

        if verbose and silent:
            raise ValueError(f"{verbose=} and {silent=} cannot both be True.")

        self.gb = None
        self.model = None
        self.var_names = None
        self.node_names = None
        self.vars = None
        self.nodes = None

    def __enter__(self):
        verbose = self.verbose

        gb = GraphBuilder(to_float32=self.to_float32).add(*self.vars_and_nodes)
        nodes, _vars = gb._all_nodes_and_vars()

        automatically_set_names = gb._set_missing_names()
        var_names = automatically_set_names["vars"]
        node_names = automatically_set_names["nodes"]

        if verbose and not self.silent:
            if var_names:
                names_ = f"The automatically assigned names are: {var_names}. "
                logger.info(f"Unnamed variables were temporarily named. {names_}")
            if node_names:
                names_ = f"The automatically assigned names are: {node_names}. "
                logger.info(f"Unnamed nodes were temporarily named. {names_}")
        elif not self.silent:
            if var_names or node_names:
                logger.info("Unnamed variables and/or nodes were temporarily named.")

        model = gb.build_model()

        self.gb = gb
        self.model = model
        self.var_names = var_names
        self.node_names = node_names
        self.vars = _vars
        self.nodes = nodes
        return model

    def __exit__(self, exc_type, exc_value, traceback):
        self.model.pop_nodes_and_vars()

        vars_dict = {var_.name: var_ for var_ in self.vars}
        nodes_dict = {node.name: node for node in self.nodes}

        for name in self.var_names:
            vars_dict[name].name = ""

        for name in self.node_names:
            nodes_dict[name].name = ""

        self.gb.nodes.clear()
        self.gb.vars.clear()

        return False  # Returning False means exceptions are not suppressed



import typing

import jax
import jax.numpy as jnp
import pytest
import tensorflow_probability.substrates.jax as tfp

import liesel.model as lsl
from liesel.distributions.nodist import NoDistribution


def test_initialization() -> None:
    # simple data variable
    var0 = lsl.Var(0, None, "")
    assert var0.value == 0
    assert var0.dist_node is None
    assert var0.name == ""
    assert isinstance(var0.value_node, lsl.Value)

    # simple data variable with specified nodes
    dat = lsl.Value(1)
    dist = lsl.Dist(NoDistribution())
    var1 = lsl.Var(dat, dist, "foo")
    assert id(var1.value_node) == id(dat)
    assert id(var1.dist_node) == id(dist)
    assert var1.value == 1
    assert var1.name == "foo"


def test_default_values() -> None:
    # simple data variable
    var = lsl.Var(0, None)
    assert var.name == ""
    assert var.dist_node is None
    assert var.role == ""
    assert var.parameter is False
    assert var.observed is False
    assert not var.groups
    assert not var.info
    assert not var.model


def test_name() -> None:
    var = lsl.Var(0.0, None, "foo")

    assert var.name == "foo"

    var.name = "bar"
    assert var.name == "bar"

    with pytest.raises(RuntimeError):
        _ = lsl.Model([var])
        var.name = "foo"


def test_role() -> None:
    var = lsl.Var(0.0, None)

    assert var.role == ""

    var.role = "bar"
    assert var.role == "bar"

    _ = lsl.Model([var])
    var.role = "foo"


def test_observed() -> None:
    var = lsl.Var(0.0, None)

    assert not var.observed

    var.observed = True
    assert var.observed

    with pytest.raises(RuntimeError):
        _ = lsl.Model([var])
        var.observed = False


def test_parameter() -> None:
    var = lsl.Var(0.0, None)

    assert not var.parameter

    var.parameter = True
    assert var.parameter

    with pytest.raises(RuntimeError):
        _ = lsl.Model([var])
        var.parameter = False


def test_info() -> None:
    var = lsl.Var(0.0, None)

    info0 = {"foo": 1.0, "bar": "baz"}
    info1 = {"foo": 1.0}

    assert not var.info

    var.info = info0
    assert var.info is info0
    assert var.info is not info1

    # infos are always changable
    _ = lsl.Model([var])
    var.info = info1
    assert var.info is info1
    assert var.info is not info0


def test_value_node() -> None:
    node0 = lsl.Value(0.0)
    node1 = lsl.Value(0.0)
    var = lsl.Var(node0, None)

    assert var.value_node is node0
    assert var.value_node is not node1

    var.value_node = node1
    assert var.value_node is not node0
    assert var.value_node is node1

    with pytest.raises(RuntimeError):
        _ = lsl.Model([var])
        var.value_node = node0


def test_dist_node() -> None:
    node0 = lsl.Dist(NoDistribution)
    node1 = lsl.Dist(NoDistribution)
    var = lsl.Var(0.0, node0)

    assert var.dist_node is node0
    assert var.dist_node is not node1

    var.dist_node = node1
    assert var.dist_node is not node0
    assert var.dist_node is node1

    with pytest.raises(RuntimeError):
        _ = lsl.Model([var])
        var.dist_node = node0


def test_property_strong_node() -> None:
    var = lsl.Var(0.0, None)

    assert var.strong
    assert not var.weak


def test_property_weak_node() -> None:
    in0 = lsl.Value(0.0)
    calc = lsl.Calc(lambda x: x, in0)
    var = lsl.Var(calc, None)

    assert not var.strong
    assert var.weak


@typing.no_type_check
def test_writing_weak_strong_fails() -> None:
    var = lsl.Var(0.0, None)

    with pytest.raises(AttributeError):
        var.strong = False

    with pytest.raises(AttributeError):
        var.weak = False


def test_read_value() -> None:
    # strong node
    var0 = lsl.Var(0, None)
    assert var0.value == 0

    # weak node
    in0 = lsl.Value(0)
    calc = lsl.Calc(lambda x: x + 1, in0)
    var1 = lsl.Var(calc, None)

    # this value might be not 1 since
    # var.update() was not called
    assert var1.value is calc.value


def test_write_value() -> None:
    # strong node
    var0 = lsl.Var(0, None)
    var0.value = 1
    assert var0.value == 1

    # weak node
    in0 = lsl.Value(0)
    calc = lsl.Calc(lambda x: x + 1, in0)
    var1 = lsl.Var(calc, None)
    with pytest.raises(RuntimeError):
        var1.value = 2


def test_property_model() -> None:
    var = lsl.Var(0.0)

    assert not var.model

    model = lsl.Model([var])
    assert var.model

    model.pop_nodes_and_vars()
    assert not var.model


def test_auto_transform():
    var = lsl.Var(1, name="var")
    var.auto_transform = True

    assert var.auto_transform

    var.auto_transform = False

    assert not var.auto_transform


def test_method_nodes() -> None:
    in0 = lsl.Value(0)
    calc = lsl.Calc(lambda x: x + 1, in0)
    dist = lsl.Dist(NoDistribution())
    var0 = lsl.Var(calc, dist)

    assert len(var0.nodes) == 3
    assert calc in var0.nodes
    assert dist in var0.nodes

    var1 = lsl.Var(in0, None)
    assert len(var1.nodes) == 2
    assert in0 in var1.nodes
    assert calc not in var1.nodes
    assert dist not in var1.nodes


def test_update_value_strong():
    var = lsl.Var(1)
    var.update()
    assert var.value == 1


def test_update_value_unfrozen_strong():
    var = lsl.Var(1)
    var.update()


def test_update_value_weak():
    var0 = lsl.Var(1, name="in")
    var1 = lsl.Var(lsl.Calc(lambda x: x + 1, var0.value_node), name="out")

    var1.update()
    assert var1.value == 2


def test_update_value_unfrozen_weak():
    var0 = lsl.Var(1, name="in")
    var1 = lsl.Var(lsl.Calc(lambda x: x + 1, var0.value_node), name="out")
    var1.update()


# ------------- test all_inputs_* / all_outputs_* ------------------------


def test_all_input_nodes_strong_no_dist():
    var = lsl.Var(0)
    assert len(var.all_input_nodes()) == 0


def test_all_input_nodes_weak_no_dist():
    x = lsl.Value(1)
    var = lsl.Var(lsl.Calc(lambda x: x + 1, x))
    assert len(var.all_input_nodes()) == 1


def test_all_input_nodes_weak_no_dist_2():
    x = lsl.Value(1)
    var = lsl.Var(
        lsl.Calc(
            lambda x, y: x + y,
            x,
            x,
        )
    )
    assert len(var.all_input_nodes()) == 1


def test_all_input_nodes_weak_no_dist_3():
    x = lsl.Value(1)
    y = lsl.Value(2)
    var = lsl.Var(
        lsl.Calc(
            lambda x, y: x + y,
            x,
            y,
        )
    )
    assert len(var.all_input_nodes()) == 2


def test_all_input_nodes_strong_w_dist():
    dist = lsl.Dist(tfp.distributions.Normal, loc=0.0, scale=1.0)

    var = lsl.Var(0.0, dist)
    assert len(var.all_input_nodes()) == 3


def test_all_input_nodes_weak_w_dist():
    dist = lsl.Dist(tfp.distributions.Normal, loc=0.0, scale=1.0)
    x = lsl.Value(1)
    var = lsl.Var(lsl.Calc(lambda x: x + 1, x), dist)
    assert len(var.all_input_nodes()) == 4


def test_all_input_nodes_weak_w_dist_2():
    dist = lsl.Dist(tfp.distributions.Normal, loc=0.0, scale=1.0)

    x = lsl.Value(1)
    var = lsl.Var(
        lsl.Calc(
            lambda x, y: x + y,
            x,
            x,
        ),
        dist,
    )
    assert len(var.all_input_nodes()) == 4


def test_all_input_nodes_weak_w_dist_3():
    dist = lsl.Dist(tfp.distributions.Normal, loc=0.0, scale=1.0)
    x = lsl.Value(1)
    y = lsl.Value(2)
    var = lsl.Var(
        lsl.Calc(
            lambda x, y: x + y,
            x,
            y,
        ),
        dist,
    )
    assert len(var.all_input_nodes()) == 5


###


def test_all_input_vars_strong_no_dist():
    var = lsl.Var(0)
    assert len(var.all_input_vars()) == 0


def test_all_input_vars_weak_no_dist():
    x = lsl.Var(1)
    var = lsl.Var(lsl.Calc(lambda x: x + 1, x))
    assert len(var.all_input_vars()) == 1


def test_all_input_vars_weak_no_dist_2():
    x = lsl.Var(1)
    var = lsl.Var(
        lsl.Calc(
            lambda x, y: x + y,
            x,
            x,
        )
    )
    assert len(var.all_input_vars()) == 1


def test_all_input_vars_weak_no_dist_3():
    x = lsl.Var(1)
    y = lsl.Var(2)
    var = lsl.Var(
        lsl.Calc(
            lambda x, y: x + y,
            x,
            y,
        )
    )
    assert len(var.all_input_vars()) == 2


def test_all_input_vars_strong_w_dist():
    dist = lsl.Dist(tfp.distributions.Normal, loc=lsl.Var(0.0), scale=lsl.Var(1.0))

    var = lsl.Var(lsl.Var(0.0), dist)
    assert len(var.all_input_vars()) == 3


def test_all_input_vars_weak_w_dist_1():
    def dist_mk():
        return lsl.Dist(tfp.distributions.Normal, loc=lsl.Var(0.0), scale=lsl.Var(1.0))

    x = lsl.Var(1)
    y = lsl.Var(1)
    z_node = lsl.Value(1)
    var = lsl.Var(lsl.Calc(lambda x: x + 1, x), dist_mk())
    assert len(var.all_input_vars()) == 3

    var = lsl.Var(lsl.Calc(lambda x, y: x + y, x, x), dist_mk())
    assert len(var.all_input_vars()) == 3

    var = lsl.Var(lsl.Calc(lambda x, y: x + y, x, y), dist_mk())
    assert len(var.all_input_vars()) == 4

    var = lsl.Var(lsl.Calc(lambda x, y: x + y, x, z_node), dist_mk())
    assert len(var.all_input_vars()) == 3


def test_all_input_vars_weak_w_dist_2():
    def dist_mk():
        return lsl.Dist(lsl.Dist(tfp.distributions.Normal, loc=0.0, scale=1.0))

    x = lsl.Var(1)
    y = lsl.Var(1)
    z_node = lsl.Value(1)
    var = lsl.Var(lsl.Calc(lambda x: x + 1, x), dist_mk())
    assert len(var.all_input_vars()) == 1

    var = lsl.Var(lsl.Calc(lambda x, y: x + y, x, x), dist_mk())
    assert len(var.all_input_vars()) == 1

    var = lsl.Var(lsl.Calc(lambda x, y: x + y, x, y), dist_mk())
    assert len(var.all_input_vars()) == 2

    var = lsl.Var(lsl.Calc(lambda x, y: x + y, x, z_node), dist_mk())
    assert len(var.all_input_vars()) == 1


def test_all_output_vars():
    x = lsl.Var(1, name="x")

    def dist_mk():
        return lsl.Dist(tfp.distributions.Normal, loc=0.0, scale=x)

    y = lsl.Var(1, name="y")
    var0 = lsl.Var(lsl.Calc(lambda x: x + 1, x), dist_mk(), name="var0")
    mod0 = lsl.Model([var0] + [x, y], copy=True)
    assert len(mod0.vars["x"].all_output_vars()) == 1
    assert len(mod0.vars["y"].all_output_vars()) == 0
    assert len(mod0.vars["var0"].all_output_vars()) == 0

    var1 = lsl.Var(lsl.Calc(lambda x, y: x + y, x, x), dist_mk(), name="var1")
    mod1 = lsl.Model([var0, var1] + [x, y], copy=True)
    assert len(mod1.vars["x"].all_output_vars()) == 2
    assert len(mod1.vars["y"].all_output_vars()) == 0
    assert len(mod1.vars["var1"].all_output_vars()) == 0

    var2 = lsl.Var(lsl.Calc(lambda x, y: x + y, x, y), dist_mk(), name="var2")
    mod2 = lsl.Model([var0, var1, var2] + [x, y], copy=True)
    assert len(mod2.vars["x"].all_output_vars()) == 3
    assert len(mod2.vars["y"].all_output_vars()) == 1
    assert len(mod2.vars["var2"].all_output_vars()) == 0


def test_all_output_nodes():
    x = lsl.Var(1, name="x")

    def dist_mk():
        return lsl.Dist(tfp.distributions.Normal, loc=0.0, scale=x)

    y = lsl.Var(1, name="y")
    var0 = lsl.Var(lsl.Calc(lambda x: x + 1, x), dist_mk(), name="var0")
    mod0 = lsl.Model([var0] + [x, y], copy=True)
    assert len(mod0.vars["x"].all_output_nodes()) == 2
    assert len(mod0.vars["y"].all_output_nodes()) == 0
    assert (
        len(mod0.vars["var0"].all_output_nodes()) == 1 + 1
    )  # part of the _model_log_prob

    var1 = lsl.Var(lsl.Calc(lambda x, y: x + y, x, x), dist_mk(), name="var1")
    mod1 = lsl.Model([var0, var1] + [x, y], copy=True)
    assert len(mod1.vars["x"].all_output_nodes()) == 4
    assert len(mod1.vars["y"].all_output_nodes()) == 0
    assert (
        len(mod1.vars["var1"].all_output_nodes()) == 1 + 1
    )  # part of the _model_log_prob

    var2 = lsl.Var(lsl.Calc(lambda x, y: x + y, x, y), dist_mk(), name="var2")
    mod2 = lsl.Model([var0, var1, var2] + [x, y], copy=True)
    assert len(mod2.vars["x"].all_output_nodes()) == 6
    assert len(mod2.vars["y"].all_output_nodes()) == 1
    assert (
        len(mod2.vars["var2"].all_output_nodes()) == 1 + 1
    )  # part of the _model_log_prob


def test_indirect_connection() -> None:
    v0 = lsl.Var(1.0, name="v0")
    n1 = lsl.Calc(lambda x: 2.0 * x, v0, _name="n1")
    v2 = lsl.Var(lsl.Calc(lambda x: 2.0 * x, n1), name="v2")
    v3 = lsl.Var(lsl.Calc(lambda x: 2.0 * x, v2), name="v3")
    _ = lsl.Model([v3])

    # test outputs
    outputs = v0.all_output_vars()
    assert len(outputs) == 1
    assert v2 in outputs
    assert v3 not in outputs

    # test inputs
    assert len(v3.all_input_vars()) == 1
    assert len(v2.all_input_vars()) == 1
    assert len(v0.all_input_vars()) == 0


class TestVarConstructors:
    def test_new_param(self):
        loc = lsl.Var.new_param(1.0, name="loc")
        assert isinstance(loc, lsl.Var)
        assert loc.parameter
        assert loc.value_node.monitor
        assert loc.strong

        dist = lsl.Dist(tfp.distributions.Normal, 0.0, 1.0)
        loc = lsl.Var.new_param(1.0, dist, name="loc")
        assert isinstance(loc, lsl.Var)
        assert loc.parameter
        assert loc.value_node.monitor
        assert loc.strong

    def test_new_obs(self):
        loc = lsl.Var.new_obs(1.0, name="loc")
        assert isinstance(loc, lsl.Var)
        assert loc.observed
        assert loc.strong

        dist = lsl.Dist(tfp.distributions.Normal, 0.0, 1.0)
        loc = lsl.Var.new_obs(1.0, dist, name="loc")
        assert isinstance(loc, lsl.Var)
        assert loc.observed
        assert loc.strong

    def test_new_calc(self):
        loc = lsl.Var.new_calc(lambda x: x + 1.0, 1.0, name="loc")
        assert isinstance(loc, lsl.Var)
        assert loc.value == pytest.approx(2.0)
        assert loc.weak

    def test_new_calc_with_dist(self):
        loc = lsl.Var.new_calc(
            lambda x: x + 1.0,
            distribution=lsl.Dist(tfp.distributions.Normal, loc=0.0, scale=1.0),
            x=1.0,
            name="loc",
        )
        loc.update()
        assert isinstance(loc, lsl.Var)
        assert loc.value == pytest.approx(2.0)
        assert loc.weak
        assert loc.dist_node is not None
        assert loc.log_prob is not None

    def test_new_const(self):
        loc = lsl.Var.new_value(1.0, name="loc")
        assert isinstance(loc, lsl.Var)
        assert loc.strong


class TestVarPredictions:
    def test_predict(self) -> None:
        n = 10
        x = jax.random.uniform(jax.random.PRNGKey(1), (n,))
        b = 1.0
        e = jax.random.normal(jax.random.PRNGKey(2), (n,))
        y = x * b + e

        xvar = lsl.Var.new_obs(x, name="x")
        bvar = lsl.Var.new_param(jnp.array([b]), name="b")
        loc = lsl.Var.new_calc(lambda x, b: x * b, x=xvar, b=bvar, name="loc")
        scale = lsl.Var.new_param(jnp.array([1.0]), name="scale")
        scale.transform(tfp.bijectors.Exp())
        yvar = lsl.Var.new_obs(
            y, lsl.Dist(tfp.distributions.Normal, loc=loc, scale=scale), name="y"
        )

        _ = lsl.Model([yvar])

        samples = {"b": jax.random.uniform(jax.random.PRNGKey(3), (4, 7))}

        pred = loc.predict(samples)
        assert jnp.allclose(pred, x * jnp.expand_dims(samples["b"], -1))
        assert pred.shape[-1] == x.shape[-1]

        # predict at new observations with same shape
        xnew = jax.random.uniform(jax.random.PRNGKey(5), (n,))
        pred = loc.predict(samples, newdata={"x": xnew})

        assert jnp.allclose(pred, xnew * jnp.expand_dims(samples["b"], -1))
        assert pred.shape[-1] == x.shape[-1]

        # predict at new grid of observations
        xnew = jnp.linspace(0, 10)
        pred = loc.predict(samples, newdata={"x": xnew})

        assert jnp.allclose(pred, xnew * jnp.expand_dims(samples["b"], -1))
        assert pred.shape[-1] != x.shape[-1]
        assert pred.shape[-1] == xnew.shape[-1]


class TestVarSample:
    @pytest.mark.parametrize("build_model", [True, False])
    def test_sample_with_model(self, build_model) -> None:
        X = lsl.Var(
            value=tfp.distributions.Uniform(low=-1.0, high=1.0).sample(
                (100, 2), jax.random.key(3)
            ),
            name="X",
        )
        b = lsl.Var(
            value=jnp.zeros(2),
            distribution=lsl.Dist(tfp.distributions.Normal, loc=0.0, scale=1.0),
            name="b",
        )
        mu = lsl.Var.new_calc(jnp.dot, X, b, name="mu")

        sigma = lsl.Var(
            1.0,
            lsl.Dist(tfp.distributions.InverseGamma, concentration=5.0, scale=0.5),
            name="sigma",
        )
        y = lsl.Var(
            jnp.zeros(X.value.shape[0]),
            lsl.Dist(tfp.distributions.Normal, mu, sigma),
            name="y",
        )
        if build_model:
            _ = lsl.Model([y])

        samples = mu.sample(shape=(1, 100), seed=jax.random.key(2))

        assert len(samples) == 1  # because there is only 1 var in the subgraph
        assert samples["b"].shape == (1, 100, 2)  # verify correct shape

        # basic plausibility checks for sampling from the correct distribution
        # this is not a tough check though.
        assert samples["b"].mean() == pytest.approx(0.0, abs=0.1)
        assert samples["b"].std() == pytest.approx(1.0, abs=0.1)

        samples = sigma.sample(shape=(1, 100), seed=jax.random.key(2))
        assert "sigma" in samples
        assert len(samples) == 1  # because there is only 1 var in the subgraph
        assert samples["sigma"].shape == (1, 100)  # verify correct shape of sigma

        # basic plausibility checks for sampling from the correct distribution
        # this is not a tough check though.
        sigma_mean = sigma.dist_node.init_dist().mean()  # type: ignore
        sigma_std = sigma.dist_node.init_dist().stddev()  # type: ignore
        assert samples["sigma"].mean() == pytest.approx(sigma_mean, abs=0.1)
        assert samples["sigma"].std() == pytest.approx(sigma_std, abs=0.1)

        samples = y.sample(shape=(1, 100), seed=jax.random.key(2))
        assert len(samples) == 3
        assert "y" in samples
        assert "sigma" in samples
        assert "b" in samples



import pytest

from liesel.model import Group, Value, Var


class TestGroup:
    def test_init(self) -> None:
        v1 = Var(0.0, name="v1")
        v2 = Var(0.0, name="v2")

        g = Group("test", var1=v1, var2=v2)

        assert "var1" in g
        assert "var2" in g
        assert g.name in v1.groups
        assert g.name in v2.groups

    def test_getitem(self) -> None:
        v1 = Var(0.0, name="v1")
        v2 = Var(0.0, name="v2")

        g = Group("test", var1=v1, var2=v2)
        assert g["var1"] == v1
        assert g["var2"] == v2

    def test_vars(self) -> None:
        v1 = Var(0.0, name="v1")
        v2 = Var(0.0, name="v2")
        n1 = Value(0.0)

        g = Group("test", var1=v1, var2=v2, nd1=n1)

        assert len(g.vars) == 2
        assert "var1" in g.vars
        assert "var2" in g.vars

    def test_nodes(self) -> None:
        v1 = Var(0.0, name="v1")
        v2 = Var(0.0, name="v2")
        n1 = Value(0.0)

        g = Group("test", var1=v1, var2=v2, nd1=n1)

        assert len(g.nodes) == 1
        assert "nd1" in g.nodes

    def test_duplicate_name(self) -> None:
        """A node cannot be member of two groups with the same name."""
        v1 = Var(0.0, name="v1")
        _ = Group("g1", var1=v1)

        with pytest.raises(RuntimeError):
            Group("g1", var1=v1)



import pytest
import tensorflow_probability.substrates.numpy.bijectors as tfb
import tensorflow_probability.substrates.numpy.distributions as tfd

from liesel.model import Calc, InputGroup, Model


def test_input_group():
    def fn(at, dargs, bargs):
        d = tfd.Normal(*dargs.args, **dargs.kwargs)
        b = tfb.Softplus(*bargs.args, **bargs.kwargs)
        td = tfd.TransformedDistribution(d, b)
        return td.log_prob(at)

    c = Calc(
        fn,
        1.0,
        dargs=InputGroup(0.0, 1.0),
        bargs=InputGroup(hinge_softness=1.0),
        _name="c",
    )

    assert c.update().value == pytest.approx(-0.6067796945571899)

    m = Model([c])
    m.update()

    assert m.nodes["c"].value == pytest.approx(-0.6067796945571899)



import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp

import liesel.goose as gs
import liesel.model as lsl


def test_custom_log_prob():
    node = lsl.Var.new_param(0.0, lsl.Dist(tfp.distributions.Normal, 0.0, 1.0), "node")
    log_prob_node = lsl.Calc(lambda lp: lp + 1.0, node.dist_node)
    mb = lsl.GraphBuilder()
    mb.add(node)
    mb.log_prob_node = log_prob_node

    model = mb.build_model()

    state = model.state
    interface = gs.LieselInterface(model)

    updated_state = interface.update_state(gs.Position({"node": 1.0}), state)

    model_log_probs = jnp.array(
        [interface.log_prob(state), interface.log_prob(updated_state)]
    )

    log_probs = tfp.distributions.Normal(0.0, 1.0).log_prob(jnp.array([0.0, 1.0])) + 1.0

    assert jnp.allclose(model_log_probs, log_probs)


def test_custom_log_prior():
    node = lsl.Var.new_param(0.0, lsl.Dist(tfp.distributions.Normal, 0.0, 1.0), "node")
    log_prob_node = lsl.Calc(lambda lp: lp + 1.0, node.dist_node)
    mb = lsl.GraphBuilder()
    mb.add(node)
    mb.log_prior_node = log_prob_node

    model = mb.build_model()

    state = model.state
    interface = gs.LieselInterface(model)

    updated_state = interface.update_state(gs.Position({"node": 1.0}), state)

    # interface does not give access to prior, so it is directly accessed using the
    # reserved node name
    model_log_priors = jnp.array(
        [state["_model_log_prior"].value, updated_state["_model_log_prior"].value]
    )

    log_probs = tfp.distributions.Normal(0.0, 1.0).log_prob(jnp.array([0.0, 1.0])) + 1.0

    assert jnp.allclose(model_log_priors, log_probs)


def test_custom_log_likelihood():
    node = lsl.Var.new_param(0.0, lsl.Dist(tfp.distributions.Normal, 0.0, 1.0), "node")
    log_prob_node = lsl.Calc(lambda lp: lp + 1.0, node.dist_node)
    mb = lsl.GraphBuilder()
    mb.add(node)
    mb.log_lik_node = log_prob_node

    model = mb.build_model()

    state = model.state
    interface = gs.LieselInterface(model)

    updated_state = interface.update_state(gs.Position({"node": 1.0}), state)

    # interface does not give access to likelihood, so it is directly accessed
    # using the reserved node name
    model_log_liks = jnp.array(
        [state["_model_log_lik"].value, updated_state["_model_log_lik"].value]
    )

    log_probs = tfp.distributions.Normal(0.0, 1.0).log_prob(jnp.array([0.0, 1.0])) + 1.0

    assert jnp.allclose(model_log_liks, log_probs)



"""
# Simple linear regression test case
"""

from collections.abc import Sequence

import jax
import jax.numpy as jnp
import numpy as np
import scipy
from pytest import approx

import liesel.goose as gs
from liesel.goose.types import Kernel

rng = np.random.default_rng(1337)

n = 30
p = 2

beta = np.ones(p)
sigma = 0.1

X = np.column_stack([np.ones(n), rng.uniform(size=[n, p - 1])])
y = rng.normal(X @ beta, sigma, size=n)

beta_ols, rss_ols, _, _ = scipy.linalg.lstsq(X, y)
sigma_ols = np.sqrt(rss_ols / (n - p))

model_state = {"y": y, "X": X, "beta": beta, "log_sigma": np.log(sigma)}


def log_prob(model_state):
    y = model_state["y"]
    mu = model_state["X"] @ model_state["beta"]
    sigma = jnp.exp(model_state["log_sigma"])

    log_probs = jax.scipy.stats.norm.logpdf(y, mu, sigma)

    return jnp.sum(log_probs)


def run_kernel_test(mcmc_seed: int, kernels: Sequence[Kernel]) -> None:
    builder = gs.EngineBuilder(mcmc_seed, num_chains=1)

    for kernel in kernels:
        builder.add_kernel(kernel)

    model = gs.DictInterface(log_prob)
    builder.set_model(model)

    builder.set_initial_values(model_state)

    builder.set_duration(
        warmup_duration=5000,
        posterior_duration=5000,
        term_duration=4000,
    )

    engine = builder.build()
    engine.sample_all_epochs()

    results = engine.get_results()
    infos = results.get_posterior_transition_infos()
    samples = results.get_posterior_samples()

    avg_beta = np.mean(samples["beta"], axis=(0, 1))
    assert avg_beta == approx(beta_ols, rel=0.05)

    avg_log_sigma = np.mean(samples["log_sigma"])
    assert avg_log_sigma == approx(np.log(sigma_ols), rel=0.05)

    for kernel in kernels:
        if hasattr(kernel, "da_target_accept"):
            avg_acceptance_prob = np.mean(infos[kernel.identifier].acceptance_prob)

            # the next line is safe since this if block checks for the existance
            # of da_target_accept
            target_accept = kernel.da_target_accept  # type: ignore
            assert avg_acceptance_prob == approx(target_accept, abs=0.05)



import tempfile

import jax.numpy as jnp
import jax.random as rnd
import matplotlib
import pytest
import tensorflow_probability.substrates.jax.distributions as tfd

from liesel.model.model import Model
from liesel.model.nodes import Dist, Var
from liesel.model.viz import plot_nodes, plot_vars

matplotlib.use("template")
key = rnd.PRNGKey(13)

n = 500
true_beta = jnp.array([1.0, 2.0])
true_sigma = 1.0

key_x, key_y = rnd.split(key, 2)
x0 = tfd.Uniform().sample(seed=key_x, sample_shape=n)
x = jnp.column_stack([jnp.ones(n), x0])

y = tfd.Normal(loc=x @ true_beta, scale=true_sigma).sample(seed=key_y)

beta_loc = Var(0.0, name="beta_loc")
beta_scale = Var(100.0, name="beta_scale")
beta_prior = Dist(tfd.Normal, loc=beta_loc, scale=beta_scale)
beta = Var(jnp.array([0.0, 0.0]), distribution=beta_prior, name="beta")

sigma_concentration = Var(0.01, name="sigma_concentration")
sigma_scale = Var(0.01, name="sigma_scale")

sigma_prior = Dist(
    tfd.InverseGamma,
    concentration=sigma_concentration,
    scale=sigma_scale,
)

sigma = Var(10.0, distribution=sigma_prior, name="sigma")

x = Var(x, name="x")
y_loc = Var.new_calc(lambda x, beta: x @ beta, x, beta, name="mu")
likelihood = Dist(tfd.Normal, loc=y_loc, scale=sigma)
y = Var(y, distribution=likelihood, name="y")

model = Model([y])


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Test plot_nodes() ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_nodes() -> None:
    fp = tempfile.TemporaryFile()
    plot_nodes(model, save_path=fp)
    matplotlib.pyplot.close()
    fp.close()


def test_plot_nodes_negative_width() -> None:
    fp = tempfile.TemporaryFile()

    with pytest.raises(ValueError):
        plot_nodes(model, width=-1, save_path=fp)

    matplotlib.pyplot.close()
    fp.close()


def test_plot_nodes_negative_height() -> None:
    fp = tempfile.TemporaryFile()

    with pytest.raises(ValueError):
        plot_nodes(model, height=-1, save_path=fp)

    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_nodes_circo_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_nodes(model, save_path=fp, prog="circo")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_nodes_dot_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_nodes(model, save_path=fp, prog="dot")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_nodes_fdp_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_nodes(model, save_path=fp, prog="fdp")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_nodes_neato_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_nodes(model, save_path=fp, prog="neato")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_nodes_osage_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_nodes(model, save_path=fp, prog="osage")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_nodes_patchwork_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_nodes(model, save_path=fp, prog="patchwork")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_nodes_twopi_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_nodes(model, save_path=fp, prog="twopi")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_nodes_sfdp_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_nodes(model, save_path=fp, prog="sfdp")
    matplotlib.pyplot.close()
    fp.close()


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Test plot_vars() ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_vars() -> None:
    fp = tempfile.TemporaryFile()
    plot_vars(model, save_path=fp)
    matplotlib.pyplot.close()
    fp.close()


def test_plot_vars_negative_width() -> None:
    fp = tempfile.TemporaryFile()

    with pytest.raises(ValueError):
        plot_vars(model, width=-1, save_path=fp)

    matplotlib.pyplot.close()
    fp.close()


def test_plot_vars_negative_height() -> None:
    fp = tempfile.TemporaryFile()

    with pytest.raises(ValueError):
        plot_vars(model, height=-1, save_path=fp)

    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_vars_circo_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_vars(model, save_path=fp, prog="circo")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_vars_dot_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_vars(model, save_path=fp, prog="dot")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_vars_fdp_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_vars(model, save_path=fp, prog="fdp")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_vars_neato_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_vars(model, save_path=fp, prog="neato")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_vars_osage_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_vars(model, save_path=fp, prog="osage")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_vars_patchwork_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_vars(model, save_path=fp, prog="patchwork")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_vars_twopi_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_vars(model, save_path=fp, prog="twopi")
    matplotlib.pyplot.close()
    fp.close()


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_vars_sfdp_prog() -> None:
    fp = tempfile.TemporaryFile()
    plot_vars(model, save_path=fp, prog="sfdp")
    matplotlib.pyplot.close()
    fp.close()



"""
This conftest.py provides default imports for doctests. For the main conftest.py,
see tests/conftest.py.
"""

import jax
import jax.numpy as jnp
import numpy as np
import pytest
import tensorflow_probability.substrates.jax.distributions as tfd

import liesel.goose as gs
import liesel.model as lsl


@pytest.fixture(autouse=True)
def add_doctest_imports(doctest_namespace):
    doctest_namespace["np"] = np
    doctest_namespace["jax"] = jax
    doctest_namespace["jnp"] = jnp
    doctest_namespace["gs"] = gs
    doctest_namespace["lsl"] = lsl
    doctest_namespace["tfd"] = tfd



import jax.numpy as jnp
import pytest
import tensorflow_probability.substrates.jax.bijectors as jb
import tensorflow_probability.substrates.jax.distributions as tfd

from liesel.model.legacy import (
    PIT,
    Addition,
    Bijector,
    ColumnStack,
    DesignMatrix,
    Hyperparameter,
    InverseLink,
    Parameter,
    Predictor,
    RegressionCoef,
    Response,
    Smooth,
    SmoothingParam,
)
from liesel.model.nodes import Dist, Value, Var


def test_design_matrix() -> None:
    x = DesignMatrix(value=jnp.eye(2), name="x")
    assert jnp.allclose(x.value, jnp.eye(2))


def test_hyperparameter() -> None:
    x = Hyperparameter(value=1.0, name="x")
    assert jnp.allclose(x.value, 1.0)


def test_parameter() -> None:
    loc = Var.new_param(0.0, name="loc")
    scale = Var.new_param(1.0, name="scale")
    dist = Dist(tfd.Normal, loc, scale)
    x = Parameter(value=1.0, distribution=dist, name="x")
    x.update()

    assert jnp.allclose(x.value, 1.0)
    assert jnp.allclose(x.log_prob, tfd.Normal(0.0, 1.0).log_prob(1.0))


def test_regression_coef() -> None:
    arr = jnp.array([1.0, 3.0, 2.3])

    loc = Var.new_param(0.0, name="loc")
    scale = Var.new_param(1.0, name="scale")
    dist = Dist(tfd.Normal, loc, scale)
    beta = RegressionCoef(value=arr, distribution=dist, name="beta")
    beta.update()

    assert jnp.allclose(beta.value, arr)
    assert jnp.allclose(beta.log_prob, tfd.Normal(0.0, 1.0).log_prob(arr))


def test_response() -> None:
    arr = jnp.array([1.0, 0.0, 2.0])

    loc = Var.new_param(0.0, name="loc")
    scale = Var.new_param(1.0, name="scale")
    dist = Dist(tfd.Normal, loc, scale)
    y = Response(value=arr, distribution=dist, name="y")
    y.update()

    assert jnp.allclose(y.value, arr)
    assert jnp.allclose(y.log_prob, tfd.Normal(0.0, 1.0).log_prob(arr))


def test_smoothing_param() -> None:
    loc = Var.new_param(0.0, name="loc")
    scale = Var.new_param(1.0, name="scale")
    dist = Dist(tfd.Normal, loc, scale)
    tau2 = SmoothingParam(value=1.0, distribution=dist, name="tau2")
    tau2.update()

    assert jnp.allclose(tau2.value, 1.0)
    assert jnp.allclose(tau2.log_prob, tfd.Normal(0.0, 1.0).log_prob(1.0))


def test_addition() -> None:
    a = Value(1.0, _name="a")
    b = Value(2.0, _name="b")
    addition = Addition(a, b)
    addition.update()

    assert jnp.allclose(addition.value, 3.0)


def test_predictor() -> None:
    a = Value(1.0, _name="a")
    b = Value(2.0, _name="b")
    predictor = Predictor(a, b)
    predictor.update()

    assert jnp.allclose(predictor.value, 3.0)


def test_bijector() -> None:
    a = Value(1.0, _name="a")
    exp_bijector = Bijector(a, jb.Exp)
    exp_bijector.update()

    assert jnp.allclose(exp_bijector.value, jnp.e)

    log_bijector = Bijector(a, jb.Exp, inverse=True)
    log_bijector.update()

    assert jnp.allclose(log_bijector.value, 0.0)


def test_inverse_link() -> None:
    a = Value(0.5, _name="a")
    log_inverse_link = InverseLink(a, jb.Log)
    log_inverse_link.update()

    assert jnp.allclose(log_inverse_link.value, jnp.log(0.5))


def test_column_stack() -> None:
    a = Value(13.0, _name="a")
    b = Value(73.0, _name="b")
    c = Value(10.0, _name="c")

    cs = ColumnStack(a, b, c)
    cs.update()

    assert jnp.allclose(cs.value, jnp.array([13.0, 73.0, 10.0]))


def test_column_stack_with_distribution() -> None:
    a = Value(1.0, _name="a")
    b = Value(0.5, _name="b")
    c = Value(0.0, _name="c")

    loc = Value(jnp.zeros(3), _name="loc")
    cov = Value(jnp.eye(3), _name="cov")
    dist = Dist(tfd.MultivariateNormalFullCovariance, loc, cov)

    cs = ColumnStack(a, b, c, distribution=dist)
    cs.update()

    tfp_dist = tfd.MultivariateNormalFullCovariance(jnp.zeros(3), jnp.eye(3))
    assert jnp.allclose(cs.log_prob, tfp_dist.log_prob([1.0, 0.5, 0.0]))


def test_pit_var() -> None:
    x = Var(1.0, name="x")

    with pytest.raises(RuntimeError):
        pit = PIT(x)

    loc = Value(0.0, _name="loc")
    scale = Value(1.0, _name="scale")
    dist = Dist(tfd.Normal, loc, scale)
    x.dist_node = dist

    pit = PIT(x)
    pit.update()

    assert jnp.allclose(pit.value, tfd.Normal(0.0, 1.0).cdf(1.0))


def test_pit_dist() -> None:
    loc = Value(0.0, _name="loc")
    scale = Value(1.0, _name="scale")
    dist = Dist(tfd.Normal, loc, scale)
    dist.at = Value(1.0, _name="x")

    pit = PIT(dist)
    pit.update()

    assert jnp.allclose(pit.value, tfd.Normal(0.0, 1.0).cdf(1.0))


def test_smooth() -> None:
    x = DesignMatrix(value=jnp.eye(2), name="x")
    beta = SmoothingParam(value=jnp.array([1.0, 3.0]), name="beta")
    smooth = Smooth(x, beta)
    smooth.update()

    assert jnp.allclose(smooth.value, jnp.array([1.0, 3.0]))



"""
tests of liesel.model.Var concerning log_prob and transform
"""

import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp

import liesel.model.nodes as lnodes


def test_log_prob_no_dist() -> None:
    var = lnodes.Var(0.0, None)
    assert var.log_prob == 0.0

    var.update()
    assert var.log_prob == 0.0


def test_log_prob_std_normal() -> None:
    dist = lnodes.Dist(tfp.distributions.Normal, 0.0, 1.0)
    var = lnodes.Var(0.0, dist)
    var.update()
    assert var.log_prob == tfp.distributions.Normal(0.0, 1.0).log_prob(0.0)

    dist = lnodes.Dist(tfp.distributions.Normal, 0.0, 1.0)
    var = lnodes.Var(1.0, dist)
    var.update()
    assert var.log_prob == tfp.distributions.Normal(0.0, 1.0).log_prob(1.0)


def test_log_prob_std_normal_vector() -> None:
    dist = lnodes.Dist(tfp.distributions.Normal, 0.0, 1.0)
    var = lnodes.Var(jnp.array([0.0, 0, 0]), dist)
    var.update()
    assert jnp.allclose(
        var.log_prob, tfp.distributions.Normal(0.0, 1.0).log_prob([0.0, 0, 0])
    )



from collections.abc import Iterator

import jax
import jax.numpy as jnp
import numpy as np
import pytest
import tensorflow_probability.substrates.jax.distributions as tfd

import liesel.goose as gs
import liesel.model as lsl
from liesel.goose.optim import (
    Stopper,
    _find_observed,
    _generate_batch_indices,
    _validate_log_prob_decomposition,
    history_to_df,
    optim_flat,
)
from liesel.goose.types import Array

key = jax.random.PRNGKey(42)
key, subkey = jax.random.split(key)
target_params = 0.5

# Generate some data.
xs = jax.random.normal(key, (100, 2))
ys = jnp.sum(xs * target_params, axis=-1) + jax.random.normal(subkey, (100,))

ols_coef = np.linalg.inv(xs.T @ xs) @ xs.T @ ys


def setup_model(ys: Array, xs: Array) -> lsl.Model:
    x = lsl.Var.new_obs(xs, name="x")
    coef = lsl.Var.new_param(jnp.zeros(2), name="coef")
    mu = lsl.Var.new_calc(jnp.dot, x, coef, name="mu")
    log_sigma = lsl.Var(2.0, name="log_sigma")
    sigma = lsl.Var.new_calc(jnp.exp, log_sigma, name="sigma")

    ydist = lsl.Dist(tfd.Normal, loc=mu, scale=sigma)
    y = lsl.Var.new_obs(ys, ydist, name="y")

    model = lsl.Model([y])
    return model


def train_test_split(model_fn, test_share: float = 0.20, **model_args):
    train = dict()
    test = dict()
    for name, value in model_args.items():
        train_idx = int(len(value) * (1 - test_share)) + 1
        test_idx = int(len(value) * test_share)

        train[name] = value[:train_idx]
        test[name] = value[-test_idx:]

    return model_fn(**train), model_fn(**test)


@pytest.fixture
def models() -> Iterator[tuple[lsl.Model, lsl.Model]]:
    yield train_test_split(setup_model, xs=xs, ys=ys)


class TestOptim:
    def test_optim_flat_jointly(self, models):
        model, _ = models

        xs = model.vars["x"].value
        ys = model.vars["y"].value
        ols_coef = np.linalg.inv(xs.T @ xs) @ xs.T @ ys
        n = ys.shape[-1]
        ols_log_sigma = jnp.log(jnp.sqrt(jnp.square(ys - (xs @ ols_coef)).sum() / n))

        stopper = Stopper(max_iter=1_000, patience=30)
        result = optim_flat(
            model, ["coef", "log_sigma"], batch_size=None, stopper=stopper
        )
        assert jnp.allclose(result.position["coef"], ols_coef, atol=1e-3)
        assert jnp.allclose(result.position["log_sigma"], ols_log_sigma, atol=1e-3)

    def test_optim_no_early_stop(self, models):
        model, _ = models

        stopper = Stopper(max_iter=1_000, patience=1_000)
        result = optim_flat(
            model, ["coef", "log_sigma"], batch_size=None, stopper=stopper
        )

        assert result.iteration == 999
        assert result.iteration_best < result.iteration
        assert result.history["loss_train"].shape == (1000,)
        assert result.history["loss_validation"].shape == (1000,)

    def test_optim_flat_batched(self, models):
        model, _ = models

        stopper = Stopper(max_iter=1000, patience=30)
        result_batched_small = optim_flat(
            model,
            ["coef", "log_sigma"],
            batch_size=10,
            batch_seed=1,
            stopper=stopper,
            model_validation=model,
        )
        result_batched_big = optim_flat(
            model,
            ["coef", "log_sigma"],
            batch_size=40,
            batch_seed=1,
            stopper=stopper,
            model_validation=model,
        )
        result_nonbatched = optim_flat(
            model, ["coef", "log_sigma"], stopper=stopper, model_validation=model
        )

        assert result_batched_small.iteration != result_nonbatched.iteration
        assert result_batched_small.iteration != result_batched_big.iteration
        assert result_nonbatched.iteration != result_batched_big.iteration

    def test_optim_flat_train_validation(self, models):
        model, model_validation = models
        stopper = Stopper(max_iter=1_000, patience=30)

        result_train = optim_flat(
            model, ["coef", "log_sigma"], batch_size=None, stopper=stopper
        )

        result_train_validation = optim_flat(
            model,
            ["coef", "log_sigma"],
            batch_size=None,
            stopper=stopper,
            model_validation=model_validation,
        )

        assert result_train.n_train == result_train.n_validation
        assert result_train_validation.n_train != result_train_validation.n_validation

        assert not jnp.allclose(
            result_train_validation.history["loss_train"],
            result_train_validation.history["loss_validation"],
        )
        assert jnp.allclose(
            result_train.history["loss_train"], result_train.history["loss_validation"]
        )

    def test_track_keys(self, models):
        model, _ = models
        stopper = Stopper(max_iter=10, patience=1)

        result = optim_flat(
            model,
            params=["coef", "log_sigma"],
            batch_size=None,
            stopper=stopper,
            track_keys=["sigma"],
        )

        assert "sigma" in result.history["tracked"]
        assert jnp.allclose(
            result.history["tracked"]["sigma"],
            jnp.exp(result.history["position"]["log_sigma"]),
        )


class TestLogProbDecompositionValidation:
    def test_const_priors(self):
        x = lsl.Var.new_obs(xs, name="x")
        coef = lsl.Var.new_param(jnp.zeros(2), name="coef")
        mu = lsl.Var.new_calc(jnp.dot, x, coef, name="mu")
        log_sigma = lsl.Var(2.0, name="log_sigma")
        sigma = lsl.Var.new_calc(jnp.exp, log_sigma, name="sigma")

        ydist = lsl.Dist(tfd.Normal, loc=mu, scale=sigma)
        y = lsl.Var.new_obs(ys, ydist, name="y")

        model = lsl.Model([y])

        interface = gs.LieselInterface(model)
        position = interface.extract_position(["coef", "log_sigma"], model.state)

        assert _validate_log_prob_decomposition(interface, position, model.state)

    def test_prior(self):
        x = lsl.Var.new_obs(xs, name="x")
        coef = lsl.Var.new_param(
            jnp.zeros(2),
            distribution=lsl.Dist(tfd.Normal, loc=0.0, scale=10.0),
            name="coef",
        )
        mu = lsl.Var.new_calc(jnp.dot, x, coef, name="mu")
        log_sigma = lsl.Var(2.0, name="log_sigma")
        sigma = lsl.Var.new_calc(jnp.exp, log_sigma, name="sigma")

        ydist = lsl.Dist(tfd.Normal, loc=mu, scale=sigma)
        y = lsl.Var.new_obs(ys, ydist, name="y")

        model = lsl.Model([y])

        interface = gs.LieselInterface(model)
        position = interface.extract_position(["coef", "log_sigma"], model.state)

        assert _validate_log_prob_decomposition(interface, position, model.state)

    def test_prior_but_not_param(self):
        x = lsl.Var.new_obs(xs, name="x")
        coef = lsl.Var(
            jnp.zeros(2),
            distribution=lsl.Dist(tfd.Normal, loc=0.0, scale=10.0),
            name="coef",
        )
        mu = lsl.Var.new_calc(jnp.dot, x, coef, name="mu")
        log_sigma = lsl.Var(2.0, name="log_sigma")
        sigma = lsl.Var.new_calc(jnp.exp, log_sigma, name="sigma")

        ydist = lsl.Dist(tfd.Normal, loc=mu, scale=sigma)
        y = lsl.Var.new_obs(ys, ydist, name="y")

        model = lsl.Model([y])

        interface = gs.LieselInterface(model)
        position = interface.extract_position(["coef", "log_sigma"], model.state)

        with pytest.raises(ValueError, match="cannot correctly be decomposed"):
            _validate_log_prob_decomposition(interface, position, model.state)

    def test_not_obs(self):
        x = lsl.Var.new_obs(xs, name="x")
        coef = lsl.Var.new_param(
            jnp.zeros(2),
            distribution=lsl.Dist(tfd.Normal, loc=0.0, scale=10.0),
            name="coef",
        )
        mu = lsl.Var.new_calc(jnp.dot, x, coef, name="mu")
        log_sigma = lsl.Var(2.0, name="log_sigma")
        sigma = lsl.Var.new_calc(jnp.exp, log_sigma, name="sigma")

        ydist = lsl.Dist(tfd.Normal, loc=mu, scale=sigma)
        y = lsl.Var(ys, ydist, name="y")

        model = lsl.Model([y])

        interface = gs.LieselInterface(model)
        position = interface.extract_position(["coef", "log_sigma"], model.state)

        with pytest.raises(ValueError, match="cannot correctly be decomposed"):
            _validate_log_prob_decomposition(interface, position, model.state)


def test_position_history_to_df(models):
    model, _ = models
    stopper = Stopper(max_iter=1_000, patience=10)
    result = optim_flat(
        model,
        ["coef", "log_sigma"],
        batch_size=20,
        stopper=stopper,
        batch_seed=1,
        prune_history=False,
    )
    df = history_to_df(result.history["position"])

    assert df.shape == (1000, 4)


def test_full_history_to_df(models):
    model, _ = models
    stopper = Stopper(max_iter=1_000, patience=10)
    result = optim_flat(
        model,
        ["coef", "log_sigma"],
        batch_size=20,
        stopper=stopper,
        batch_seed=1,
        prune_history=False,
        track_keys=["sigma"],
    )
    df = history_to_df(result.history)

    assert df.shape == (1000, 7)


def test_history_to_df_pruned(models):
    model, _ = models
    stopper = Stopper(max_iter=1_000, patience=10)
    with jax.disable_jit(disable=False):
        result = optim_flat(
            model,
            ["coef", "log_sigma"],
            batch_size=20,
            stopper=stopper,
            batch_seed=1,
            prune_history=True,
            model_validation=model,
        )
    df = history_to_df(result.history["position"])

    assert df.shape[0] < 100


def test_generate_batches():
    key = jax.random.PRNGKey(42)
    jit_batches = jax.jit(_generate_batch_indices, static_argnames=["n", "batch_size"])
    batches = jit_batches(key, n=30, batch_size=9)
    assert len(batches) == 3

    for batch in batches:
        assert len(batch) == 9


def test_find_observed():
    n = 10
    a = lsl.Var.new_obs(jnp.arange(n), name="a")
    model = lsl.Model([a])
    observed = _find_observed(model)

    assert "a" in observed


def test_find_observed_weak():
    n = 10
    a = lsl.Var.new_obs(jnp.arange(n), name="a")
    b_calc = lsl.Var.new_calc(lambda x: x + 1, a)
    b = lsl.Var.new_obs(b_calc, name="b")
    model = lsl.Model([b])
    observed = _find_observed(model)

    assert "a" in observed
    assert "b" not in observed


class TestStopper:
    def test_stopper_does_not_stop(self):
        stopper = Stopper(patience=5, max_iter=100)

        # case 1: continous decrease
        loss_history = np.linspace(100, 0, 100)
        stop = stopper.stop_early(i=6, loss_history=loss_history)
        assert not stop

        # case 2: current loss is not the best
        loss_history = np.zeros((100,))
        loss_history[5] = -1.0
        stop = stopper.stop_early(i=6, loss_history=loss_history)
        assert not stop

    def test_stopper_stops_atol(self):
        stopper = Stopper(patience=5, max_iter=100, atol=0.1)

        loss_history = np.zeros((100,))
        loss_history[2:7] = np.array([0.0, 1.0, -0.05, -0.05, -0.01])
        stop = stopper.stop_early(i=6, loss_history=loss_history)
        assert stop

    def test_stopper_stops_rtol(self):
        stopper = Stopper(patience=5, max_iter=100, atol=0.0, rtol=0.05)

        # case one: 0.0 gets compared to -0.05
        # relative difference is 1
        # so no stop
        loss_history = np.zeros((100,))
        loss_history[2:7] = np.array([0.0, 1.0, -0.05, -0.05, -0.01])
        stop = stopper.stop_early(i=6, loss_history=loss_history)
        assert not stop

        # case one: -0.049 gets compared to -0.05
        # relative difference is ~0.02
        # so stop
        loss_history[2:7] = np.array([-0.049, 1.0, -0.05, -0.05, -0.01])
        stop = stopper.stop_early(i=6, loss_history=loss_history)
        assert stop

    def test_stopper_jitted(self):
        stopper = Stopper(patience=5, max_iter=100, atol=0.1, rtol=0.1)
        stop_jit = jax.jit(stopper.stop_early)

        loss_history = jnp.array([1.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
        stop = stop_jit(i=6, loss_history=loss_history)
        assert stop

    def test_stop_at_jitted(self):
        stopper = Stopper(patience=5, max_iter=100, atol=0.1, rtol=0.1)
        stop_at_jit = jax.jit(stopper.which_best_in_recent_history)

        key = jax.random.PRNGKey(42)
        loss_history = 2.0 + jax.random.uniform(
            key, shape=(15,), minval=0.0, maxval=0.1
        )

        stop = stopper.stop_early(i=6, loss_history=loss_history)
        stop_at = stop_at_jit(i=6, loss_history=loss_history)

        assert stop
        assert loss_history[stop_at] == pytest.approx(jnp.min(loss_history))

    def test_stop_at_start(self):
        stopper = Stopper(max_iter=100, patience=5, atol=0.1, rtol=0.1)
        loss_history = jnp.zeros(shape=(10,))

        stop = stopper.stop_now(i=0, loss_history=loss_history)

        assert not stop

    def test_patience(self):
        stopper = Stopper(max_iter=100, patience=100)
        loss_history = jnp.zeros(shape=(100,))

        for i in range(100):
            assert not stopper.stop_early(i, loss_history)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os

# import sys
# sys.path.insert(0, os.path.abspath("."))

html_context = {}

on_rtd = os.environ.get("READTHEDOCS") == "True"

# Set canonical URL from the Read the Docs Domain
html_baseurl = os.environ.get("READTHEDOCS_CANONICAL_URL", "")

# Tell Jinja2 templates the build is running on Read the Docs
if os.environ.get("READTHEDOCS", "") == "True":
    html_context["READTHEDOCS"] = True

# -- Project information -----------------------------------------------------

project = "liesel"
copyright = "2022, Hannes Riebl & Paul Wiemann"
author = "Hannes Riebl & Paul Wiemann"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",  # parse NumPy and Google style docstrings
    "sphinx_autodoc_typehints",
    "sphinx.ext.autosummary",  # for automatic API doc tables
    # "sphinx.ext.linkcode",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "sphinx_remove_toctrees",  # speed up builds with many stub pages
    "sphinx.ext.intersphinx",
    "sphinx_copybutton",
    "myst_nb",
]

if on_rtd:
    extensions.append("rtds_action")

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy-1.8.1/", None),
    "jax": ("https://jax.readthedocs.io/en/latest/", None),
    "tensorflow_probability": (  # we can use "tfp" as a prefix, too
        "https://www.tensorflow.org/probability/api_docs/python",
        "https://github.com/GPflow/tensorflow-intersphinx/raw/master/tfp_py_objects.inv",  # noqa: E501
    ),
}

# Napoleon options
napoleon_use_param = True

# For compatibility with sphinx_autodoc_typehints:
# If True, the return text will be rendered as literals.
napoleon_preprocess_types = False

# For compatibility with sphinx_autodoc_typehints:
# If True, Napoleon will add a :rtype: role, causing sphinx_autodoc_typehints
# to not add its own role from the type annotations.
napoleon_use_rtype = False

# sphinx_autodoc_typehints options
typehints_defaults = "braces-after"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to the source directory, that match files and
# directories to ignore when looking for source files. These patterns also
# affect html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

pygments_style = "sphinx"

# The theme to use for HTML and HTML help pages. See the documentation for
# a list of builtin themes.
html_theme = "sphinx_book_theme"
# html_theme = "pydata_sphinx_theme"
html_title = ""
html_logo = "../../misc/logo/logo-light.png"
html_theme_options = {
    "repository_url": "https://github.com/liesel-devs/liesel",
    "use_repository_button": True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["custom.css"]


def linkcode_resolve(domain, info):
    """For the linkcode extension."""
    if domain != "py":
        return None
    if not info["module"]:
        return None
    filename = info["module"].replace(".", "/")
    return f"https://github.com/liesel-devs/liesel/blob/main/{filename}.py"


# Mock / ignore the following modules.
autodoc_mock_imports = ["liesel.distributions.nodist"]

# Map functions and classes with the same lowercase names to other filenames.
autosummary_filename_map = {
    "liesel.goose.summary_m.summary": "liesel.goose.summary_m.summary-function",
    "liesel.goose.summary_m.Summary": "liesel.goose.summary_m.summary-class",
}

# Only document module members that are listed in __all__ (if defined).
autosummary_ignore_module_all = False

# Remove auto-generated API docs from the sidebar. They take too long to build.
remove_from_toctrees = ["generated/liesel.*.*.*.*.rst"]


# --------------------------------------------------------------------------------------
# myst settings
# --------------------------------------------------------------------------------------

# myst configuration
myst_heading_anchors = 3  # auto-generate 3 levels of heading anchors
myst_enable_extensions = ["amsmath", "dollarmath", "html_image"]
myst_dmath_double_inline = True


# --------------------------------------------------------------------------------------
# rtds-action settings
# --------------------------------------------------------------------------------------

if on_rtd:
    # The name of your GitHub repository
    rtds_action_github_repo = "liesel-devs/liesel"

    # The path where the artifact should be extracted
    # Note: this is relative to the conf.py file!
    rtds_action_path = "tutorials/md"

    # The "prefix" used in the `upload-artifact` step of the action
    rtds_action_artifact_prefix = "tutorials-for-"

    # A GitHub personal access token is required, more info below
    rtds_action_github_token = os.environ["GITHUB_TOKEN"]

    # Whether or not to raise an error on Read the Docs if the
    # artifact containing the notebooks can't be downloaded (optional)
    rtds_action_error_if_missing = False



"""
Liesel modeling framework.
"""

from .distreg import DistRegBuilder, dist_reg_mcmc, tau2_gibbs_kernel
from .legacy import (
    PIT,
    Addition,
    Bijector,
    ColumnStack,
    DesignMatrix,
    Hyperparameter,
    InverseLink,
    Parameter,
    Predictor,
    RegressionCoef,
    Response,
    Smooth,
    SmoothingParam,
)
from .model import GraphBuilder, Model, load_model, save_model
from .nodes import (
    Array,
    Calc,
    Data,
    Dist,
    Distribution,  # TODO: Bijector?
    Group,
    InputGroup,
    Node,
    NodeState,
    TransientCalc,
    TransientDist,
    TransientIdentity,
    TransientNode,
    Value,
    Var,
)
from .viz import plot_nodes, plot_vars

__all__ = [
    "DistRegBuilder",
    "dist_reg_mcmc",
    "tau2_gibbs_kernel",
    "GooseModel",
    "PIT",
    "Addition",
    "Bijector",
    "ColumnStack",
    "DesignMatrix",
    "Hyperparameter",
    "InverseLink",
    "Parameter",
    "Predictor",
    "RegressionCoef",
    "Response",
    "Smooth",
    "SmoothingParam",
    "GraphBuilder",
    "Model",
    "load_model",
    "save_model",
    "Array",
    "Calc",
    "Data",
    "Dist",
    "Distribution",
    "Group",
    "InputGroup",
    "Node",
    "NodeState",
    "TransientCalc",
    "TransientDist",
    "TransientIdentity",
    "TransientNode",
    "Value",
    "Var",
    "plot_nodes",
    "plot_vars",
]
